# file autogenerated by scripts/examples_to_script.py
def ex_01_simple_cav():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import finesse

    finesse.configure(plotting=True)
    kat = finesse.Model()
    kat.parse(
        """
        # Add a Laser named L0 with a power of 1 W.
        l L0 P=1
        # Space attaching L0 <-> m1 with length of 0 m (default).
        s s0 L0.p1 m1.p1
        # Input mirror of cavity.
        m m1 R=0.99 T=0.01
        # Intra-cavity space with length of 1 m.
        s CAV m1.p2 m2.p1 L=1
        # End mirror of cavity.
        m m2 R=0.991 T=0.009
        # Power detectors on reflection, circulation and transmission.
        pd refl m1.p1.o
        pd circ m2.p1.i
        pd trns m2.p2.o
        """
    )
    # Scan over the detuning DOF of m1 from -180 deg to +180 deg with 400 points.
    out = kat.run("xaxis(m1.phi, lin, -180, 180, 400)")
    out.plot(logy=True)
    # Add some curvatures to our mirrors.
    kat.m1.Rc = -0.7
    kat.m2.Rc = 0.7
    # Define the cavity.
    kat.parse("cavity cavity1 source=m1.p2.o via=m2.p1.i priority=1")
    # Plot the beam trace, starting from cavity the eigenmode.
    tsy = finesse.tracing.tools.propagate_beam(
        to_node=kat.L0.p1.i, from_node=kat.m2.p1.o, direction="y"
    )
    tsy.plot()
    # Remove s0 so we can add in a lens.
    # FIXME kat.remove is not yet working, so we use a list
    # comprehension to remove s0 from the model and
    # reparse it into a new model. (See unparse below)
    kat2 = finesse.Model()
    kat2.parse("\n".join([l for l in kat.unparse().splitlines() if "s0" not in l]))
    # Add a silicon lens with an AR surface to the input coupler.
    kat2.parse(
        """
        space s0 L0.p1 m1_AR.p1 L=0.5
        mirror m1_AR R=100e-6 L=10e-6 Rc=-0.3
        space s_M1_sub portA=m1_AR.p2 portB=m1.p1 L=10e-3 nr=3.5
        """
    )
    # Plot beam trace, starting from the cavity eigenmode.
    tsy = finesse.tracing.tools.propagate_beam(
        to_node=kat2.L0.p1.i, from_node=kat2.m2.p1.o, direction="y"
    )
    tsy.plot()
    print(kat)
    print(kat.component_tree(kat.L0))
    print(kat.unparse())


def ex_02_pdh_lock():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import finesse

    finesse.configure(plotting=True)
    model = finesse.Model()
    model.parse(
        """
        laser i1 P=1 f=0
        space s1 portA=i1.p1 portB=eo1.p1
        modulator eo1 15M 0.7 order=3
        space s2 portA=eo1.p2 portB=m1.p1
        mirror m1 R=0.9 T=0.1
        space s3 portA=m1.p2 portB=m2.p1 L=1
        mirror m2 R=0.9 T=0.1
        power_detector_dc refl node=m1.p1.o
        power_detector_dc trans node=m2.p2.o
        """
    )
    import numpy as np

    for phase in np.linspace(0, 90, endpoint=True, num=6):
        model.parse(f"pd1 pdh{phase:.0f} node=m1.p1.o f=eo1.f phase={phase:.0f}")
    sol = model.run("xaxis(m2.phi, lin, -100, 100, 400)")
    sol.plot()

    def remove_demod_pds(verbose=True):
        for el in model.elements:
            if el.startswith("pdh"):
                model.remove(el)
                if verbose:
                    print(f"Removed {el}")

    # Remove the old detectors.
    remove_demod_pds()
    # Create another array of phases.
    phases = np.arange(0, 90, 1)
    for phase in phases:
        model.parse(f"pd1 pdh{phase:.0f} node=m1.p1.o f=eo1.f phase={phase:.0f}")
    # Sweep the mirror tuning across 3 points: ±1° and 0.
    out = model.run("xaxis(m2.phi, lin, -1, 1, 2)")
    # Calculate the gradient of the error signal.
    grad = []
    for phase in phases:
        var = out[f"pdh{phase}"]
        grad.append((var[0] - var[-1]) / 2)
    # Find the maximum gradient.
    grad = np.array(grad)
    idxmax = np.argmax(np.abs(grad))
    print(f"Maximum error signal gradient occurs at {phases[idxmax]} degrees")
    # Plot.
    from matplotlib import pyplot as plt

    plt.plot(phases, grad)
    plt.xlabel("Demodulation Phase (°)")
    plt.ylabel("Error Signal Gradient (arb)")
    remove_demod_pds(verbose=False)
    model.parse(
        """
    power_detector_demod_1 pdh node=m1.p1.o f=eo1.f phase=9
    lock lock_length pdh m1.phi -10 1e-8
    """
    )
    sol = model.run("xaxis(m2.phi, lin, 0, 100, 400, pre_step=run_locks())")
    sol.plot()
    import numpy as np

    import finesse

    finesse.init_plotting()
    model = finesse.Model()
    model.parse(
        """
        laser i1 P=1 f=0
        space s1 portA=i1.p1 portB=eo1.p1
        modulator eo1 15M 0.7 order=3
        space s2 portA=eo1.p2 portB=m1.p1
        mirror m1 R=0.9 T=0.1
        space s3 portA=m1.p2 portB=m2.p1 L=1
        mirror m2 R=0.9 T=0.1
        pd1 pdh_complex m1.p1.o f=eo1.f
        """
    )
    eps = 1e-9  # finite difference step size for gradient calculation
    sol = model.run(f"xaxis(m1.phi, lin, -{eps / 2}, {eps / 2}, 1, relative=True)")
    # compute the real and imaginary gradient with respect to the
    # cavity length change, then compute the complex argument (angle)
    # to get the optimial demodulation phase
    opt_demod = np.angle(
        np.gradient(sol["pdh_complex"], eps).mean(),  # take mean as we need
        deg=True,  # make sure we compute angle in degrees
    )
    print(opt_demod)
    model.parse(
        f"""
    pd1 pdh_I m1.p1.o f=eo1.f phase={opt_demod}
    pd1 pdh_Q m1.p1.o f=eo1.f phase={opt_demod + 90}
    """
    )
    sol = model.run("xaxis(m1.phi, lin, -50, 50, 200)")
    sol.plot("pdh_I", "pdh_Q")


def ex_03_near_unstable():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")


def ex_04_geometrical_params():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import numpy as np

    # Cavity parameters
    L = 0.1  # Cavity Length
    RcM2 = 0.5  # Radius of curvature of M2 mirror = 0.5 m
    g1 = 1  # g1 = 1 - L/RcM1 = 1 since RcM1 = infinity
    g2 = 1 - L / RcM2  # g2 for M2 mirror
    print("g1g2 =", g1 * g2)  # Stability check
    # Mirror Tilt:
    npoints = 200
    theta_low = -1e-6
    theta_high = 1e-6

    def geometric(mirror, theta_low, theta_high, npoints):
        if mirror == 1:
            # M1 tilt
            theta_1 = np.linspace(theta_low, theta_high, npoints + 1, endpoint=True)
            # M2 tilt
            theta_2 = 0
        elif mirror == 2:
            # M1 tilt
            theta_1 = 0
            # M2 tilt
            theta_2 = np.linspace(theta_low, theta_high, npoints + 1, endpoint=True)
        else:
            raise NotImplementedError("Mirror should be 1 or 2 and an integer")
        # Resultant translation
        ## Displacement on M1
        delta_x1 = (g2 / (1 - g1 * g2)) * L * theta_1 - (
            1 / (1 - g1 * g2)
        ) * L * theta_2
        ## Displacement on M2
        delta_x2 = (1 / (1 - g1 * g2)) * L * theta_1 - (
            g1 / (1 - g1 * g2)
        ) * L * theta_2
        ## Angular displacement of optical axis
        delta_theta = (delta_x2 - delta_x1) / L
        print("Results for analytical solutions:")
        print("∆x1: ", max(delta_x1), min(delta_x1), " m")
        print("∆x2: ", max(delta_x2), min(delta_x2), " m")
        print("∆θ: ", max(delta_theta), min(delta_theta), " m")
        return delta_x1, delta_x2, delta_theta

    import matplotlib.pyplot as plt

    import finesse
    from finesse import init_plotting

    init_plotting()
    T = 1e-6
    # Build Finesse model
    kat = finesse.Model()
    kat.parse(
        f"""
        laser laser P=1.0 f=0 phase=0
        # Mirrors
        mirror M1 R=1-M1.T T={T} L=0.0 phi=0 Rc=[inf, inf] xbeta=0 ybeta=0
        mirror M2 R=1-M1.T T=M1.T L=0.0 phi=0 Rc=[{RcM2}, {RcM2}] xbeta=0 ybeta=0
        # Spaces
        space s0 portA=laser.p1 portB=M1.p1 L=1 nr=1
        space s_cav portA=M1.p2 portB=M2.p1 L=0.1 nr=1
        # Amplitude Detectors
        amplitude_detector det00_1 node=M1.p2.i f=0 n=0 m=0
        amplitude_detector det00_2 node=M2.p1.o f=0 n=0 m=0
        amplitude_detector det10_1 node=M1.p2.i f=0 n=1 m=0
        amplitude_detector det10_2 node=M2.p1.o f=0 n=1 m=0
        # Beam Property Detectors
        beam_property_detector bp_n2 prop=0 node=M1.p2.i direction=x q_as_bp=false
        beam_property_detector bp_n3 prop=0 node=M2.p1.o direction=x q_as_bp=false
        beam_property_detector cavwaist prop=1 node=M1.p2.i direction=x q_as_bp=false
        # Photodiodes
        power_detector_dc pcirc node=M1.p2.o pdtype=none
        # Config
        cavity cavity1 source=M1.p2.o via=M2.p1.i priority=1
        lambda(1.064e-06)
        modes(maxtem=3)
        """
    )
    # Print beam tracing information
    tsy = finesse.tracing.tools.propagate_beam(
        to_node=kat.M2.p1.o, from_node=kat.M1.p2.o, direction="y"
    )
    print(tsy.table())

    def run_simulation(mirror, theta_low, theta_high, npoints):
        return process_sim(
            kat.run(
                f"""
                xaxis(
                    parameter=M{mirror:.0f}.xbeta,
                    mode=lin,
                    start={theta_low},
                    stop={theta_high},
                    steps={npoints},
                    pre_step=none,
                    post_step=none
                )
                """
            )
        )

    def process_sim(out):
        # Comparison
        acirc = np.sqrt(out["pcirc"].real)  # circulating amplitude
        # intermodal phase HG10 -> HG00
        # at mirror 1
        intermodal_phase_1 = np.angle(out["det10_1"]) - np.angle(out["det00_1"])
        # at mirror 2
        intermodal_phase_2 = np.angle(out["det10_2"]) - np.angle(out["det00_2"])
        # Real part of \power normalised HG10 mode ampltiude
        # at M1
        a_shift_1 = (np.abs(out["det10_1"]) / acirc) * np.cos(intermodal_phase_1)
        # at M2
        a_shift_2 = (np.abs(out["det10_2"]) / acirc) * np.cos(intermodal_phase_2)
        # Imaginary part of power normalised HG10 mode ampltiude
        # at M1
        a_tilt_1 = (np.abs(out["det10_1"]) / acirc) * np.sin(intermodal_phase_1)
        # at M2
        a_tilt_2 = (np.abs(out["det10_2"]) / acirc) * np.sin(intermodal_phase_2)
        # Spot position shift
        # at M1
        dx_finesse_1 = out["bp_n2"].real * a_shift_1
        # at M2
        dx_finesse_2 = out["bp_n3"].real * a_shift_2
        # Wavefront tilt at
        # M1
        dtheta_1 = a_tilt_1 * kat.lambda0 / (np.pi * out["cavwaist"].real)
        # M2
        dtheta_2 = a_tilt_2 * kat.lambda0 / (np.pi * out["cavwaist"].real)
        print("Results for finesse solutions:")
        print(
            "Intermodal_1: ",
            max(intermodal_phase_1 / np.pi),
            min(dx_finesse_1 / np.pi),
            " pi",
        )
        print(
            "Intermodal_2: ",
            max(intermodal_phase_2 / np.pi),
            min(dx_finesse_2 / np.pi),
            " pi",
        )
        print("∆x1: ", max(dx_finesse_1), min(dx_finesse_1), " m")
        print("∆x2: ", max(dx_finesse_2), min(dx_finesse_2), " m")
        print("∆θ1: ", max(dtheta_1), min(dtheta_1), " m")
        print("∆θ2: ", max(dtheta_2), min(dtheta_2), " m")
        return (
            intermodal_phase_1,
            intermodal_phase_2,
            dx_finesse_1,
            dx_finesse_2,
            dtheta_1,
            dtheta_2,
            out.x[0],
        )

    def comparison_plot(mirror, theta_low, theta_high, npoints, ax=None):
        delta_x1, delta_x2, delta_theta = geometric(
            mirror, theta_low, theta_high, npoints
        )
        (
            intermodal_phase_1,
            intermodal_phase_2,
            dx_finesse_1,
            dx_finesse_2,
            dtheta_1,
            dtheta_2,
            xaxis,
        ) = run_simulation(mirror, theta_low, theta_high, npoints)
        if ax is None:
            fig, ax = plt.subplots(nrows=3, sharex=True, figsize=(4, 6 / 3 * 4))
        xaxis = 1e6 * xaxis
        ax2col = "darkgreen"
        # Plot dx1
        (l3,) = ax[0].plot(xaxis, 1e6 * delta_x1, "-", label="Geometric")
        (l4,) = ax[0].plot(xaxis, 1e6 * dx_finesse_1, "--", label="Modal")
        ax[0].plot([], [], "--", color=ax2col, label="Difference")
        ax[0].set_ylabel(r"$\Delta \mathrm{x}_1$ [µm]")
        # Plot dx1 err
        ax1_2 = ax[0].twinx()
        (l5,) = ax1_2.plot(xaxis, 1e12 * (delta_x1 - dx_finesse_1), "--", color=ax2col)
        ax1_2.tick_params(axis="y", colors=ax2col)
        ax1_2.set_ylabel("Diff. [pm]", color=ax2col)
        # plot dx2
        ax[1].plot(xaxis, 1e6 * delta_x2, "-", label="Geometric")
        ax[1].plot(xaxis, 1e6 * dx_finesse_2, "--", label="Modal")
        ax[1].plot([], [], "--", color=ax2col, label="Difference")
        ax[1].set_ylabel(r"$\Delta \mathrm{x}_2$ [µm]")
        # plot dx2 err
        ax2_2 = ax[1].twinx()
        ax2_2.plot(xaxis, 1e12 * (delta_x2 - dx_finesse_2), "--", color=ax2col)
        ax2_2.tick_params(axis="y", colors=ax2col)
        ax2_2.set_ylabel("Diff. [pm]", color=ax2col)
        # plot dtheta1
        ax[2].plot(xaxis, 1e6 * delta_theta, "-", label="Geometric")
        ax[2].plot(xaxis, 1e6 * dtheta_1, "--", label="Modal")
        ax[2].plot([], [], "--", color=ax2col, label="Difference")
        ax[2].set_ylabel(r"$\theta^\mathrm{x}_0$ at M1 [µrad]")
        # plot dtheta1 err
        ax3_2 = ax[2].twinx()
        ax3_2.plot(xaxis, 1e12 * (dtheta_1 - delta_theta), "--", color=ax2col)
        ax3_2.tick_params(axis="y", colors=ax2col)
        ax3_2.set_ylabel("Diff. [prad]", color=ax2col)
        ax[0].legend()
        ax[2].set_xlabel(f"M{mirror} misalignment [µrad]")

    fig, ax = plt.subplots(nrows=3, ncols=2, sharex=True, figsize=(8, 6 / 3 * 4))
    comparison_plot(1, theta_low, theta_high, npoints, ax=ax[:, 0])
    comparison_plot(2, theta_low, theta_high, npoints, ax=ax[:, 1])
    ax[2, 1].set_ylim([-1.1, +1.1])
    fig.tight_layout()
    kat.M2.xbeta = theta_high
    kat.M1.xbeta = 0
    (
        intermodal_phase_1,
        intermodal_phase_2,
        dx_finesse_1,
        dx_finesse_2,
        dtheta_1,
        dtheta_2,
        xaxis,
    ) = process_sim(
        kat.run(
            f"""
            xaxis(
                parameter=M1.T,
                mode=log,
                start=1e-7,
                stop=1,
                steps={npoints},
                pre_step=none,
                post_step=none)
            """
        )
    )
    delta_x1, delta_x2, delta_theta = geometric(2, 0, kat.M2.xbeta, npoints)
    fig, ax = plt.subplots(figsize=(8, 8 / 3))
    ## Plot phases
    ax.loglog(
        xaxis, 1e6 * np.abs(delta_x1[-1] - dx_finesse_1), "-", label="Plane mirror (M1)"
    )
    ax.loglog(
        xaxis,
        1e6 * np.abs(delta_x2[-1] - dx_finesse_2),
        "--",
        label="Hemispherical mirror (M2)",
    )
    ax.set_ylabel("Difference in predicted\n spot position [µm]")
    ax.set_xlabel("Mirror transmissivity")
    ax.set_title(f"Mirror M2 tuning {1e6 * kat.M2.xbeta:.0f}urad")
    ax.legend()


def ex_05_modulation():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import finesse

    finesse.configure(plotting=True)
    kat = finesse.Model()
    kat.parse(
        """
        # Laser with power 1W
        laser L0 P=1
        s s0 L0.p1 eom1.p1
        # Phase modulator at 40kHz, with a modulation index of 0.05
        mod eom1 f=40k midx=.05 order=5 mod_type=pm
        # Three amplitude detectors, looking at the first three positive modulation
        # sidebands
        ad bessel1 eom1.p2.o f=40k
        ad bessel2 eom1.p2.o f=80k
        ad bessel3 eom1.p2.o f=120k
        """
    )
    # Sweep the modulation index of the modulator from 0 to 10
    out = kat.run("xaxis(eom1.midx, lin, 0, 10, 300)")
    out.plot()


def ex_06_radiation_pressure():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import finesse

    finesse.configure(plotting=True)
    kat = finesse.Model()
    kat.parse(
        """
        # Optical setup: laser, space and mirror:
        l l1 P=1
        s s1 l1.p1 m1.p1
        m m1 R=1 T=0
        # Define a pendulum for our mirror, with a z-motion resonance at 10Hz
        # and Q factor of 1000
        pendulum sus1 m1.mech mass=1 fz=10 Qz=1000
        # Measure the mirror's longitudinal motion
        xd m1_z m1.mech.z
        # Set a signal frequency to activate the signal simulation
        # (needed for radiation pressure effects to be observed)
        fsig(1)
        # Generate amplitude modulation at the laser
        sgen sig l1.amp
        """
    )
    # Sweep the frequency of the amplitude modulation
    out = kat.run("xaxis(fsig.f, log, 1, 100, 400)")
    out.plot(log=True)


def ex_07_homodyne():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
    import numpy as np

    import finesse

    finesse.configure(plotting=True)
    kat = finesse.Model()
    kat.parse(
        """
        l l1 P=1-bs1.T
        s s1 l1.p1 bs1.p4 L=1
        sq sq1 db=10
        s s2 sq1.p1 bs1.p1
        bs bs1 R=0.999 L=0 alpha=45
        # Set a signal frequency to activate the signal simulation
        # (needed for squeezing effects to be observed)
        fsig(1)
        # Output the pure shot noise along with the qnoised detector
        # measuring the effects of the squeezing
        qnoised sqzd_noise bs1.p2.o
        qshot shot_noise bs1.p2.o
    """
    )
    out = kat.run("xaxis(l1.phase, lin, -90, 90, 360)")
    out.plot(separate=False)

    def to_db(quantity, reference, isAmplitude=False):
        if isAmplitude:
            factor = 20
        else:
            factor = 10
        return factor * np.log10(quantity / reference)

    sqz_db = to_db(out["sqzd_noise"], out["shot_noise"], isAmplitude=True)
    fig, ax = plt.subplots()
    ax.plot(out.x0, sqz_db)
    ax.set_ylabel("SQZ / ASQZ [dB]")
    ax.set_xlabel("LO Phase [deg]")


def ex_08_optical_spring():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import finesse

    finesse.configure(plotting=True)
    kat = finesse.Model()
    kat.parse(
        """
        l l1 P=3
        s s1 l1.p1 Min.p1 L=1
        # The cavity
        m Min R=0.9937 L=0
        s scav Min.p2 Mend.p1 L=1
        m Mend R=1 L=0 phi=-0.048  # Roughly 40kHz detuning
        # Suspend mirrors on pendula
        pendulum sus1 Min.mech mass=0.25 fz=1 Qz=1000
        pendulum sus2 Mend.mech mass=0.25 fz=1 Qz=1000
        # Apply a `force signal' to the end mirror
        fsig(1)
        sgen force Mend.mech.F_z
        # we want to measure the longitudinal motion signal of both mirrors
        xd Mend_z Mend.mech.z
        xd Min_z Min.mech.z
        """
    )
    out = kat.run("xaxis(fsig.f, log, 0.1, 1k, 300)")
    out.plot(log=True)


def ex_09_aligo_sensitivity():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import finesse

    finesse.configure(plotting=True)
    kat = finesse.Model()
    kat.parse(
        """
        ###########################################################################
        ###   Variables
        ###########################################################################
        var Larm 3995
        var Mtm  40
        var itmT 0.014
        var lmichx 4.5
        var lmichy 4.45
        ###########################################################################
        ###   Input optics
        ###########################################################################
        l L0 125
        s l_in L0.p1 prm.p1
        # Power recycling mirror
        m prm T=0.03 L=37.5u phi=90
        s prc prm.p2 bs.p1 L=53
        # Central beamsplitter
        bs bs T=0.5 L=0 alpha=45
        ###########################################################################
        ###   X arm
        ###########################################################################
        s lx bs.p3 itmx.p1 L=lmichx
        m itmx T=itmT L=37.5u phi=90
        s LX itmx.p2 etmx.p1 L=Larm
        m etmx T=5u L=37.5u phi=89.999875
        pendulum itmx_sus itmx.mech mass=Mtm fz=1 Qz=1M
        pendulum etmx_sus etmx.mech mass=Mtm fz=1 Qz=1M
        ###########################################################################
        ###   Y arm
        ###########################################################################
        s ly bs.p2 itmy.p1 L=lmichy
        m itmy T=itmT L=37.5u phi=0
        s LY itmy.p2 etmy.p1 L=Larm
        m etmy T=5u L=37.5u phi=0.000125
        pendulum itmy_sus itmy.mech mass=Mtm fz=1 Qz=1M
        pendulum etmy_sus etmy.mech mass=Mtm fz=1 Qz=1M
        ###########################################################################
        ###   Output and squeezing
        ###########################################################################
        s src bs.p4 srm.p1 L=50.525
        m srm T=0.2 L=37.5u phi=-90
        # A squeezed source could be injected into the dark port
        sq sq1 db=0 angle=90
        s lsqz sq1.p1 srm.p2
        # Differentially modulate the arm lengths
        fsig(1)
        sgen darmx LX.h
        sgen darmy LY.h phase=180
        # Output the full quantum noise limited sensitivity
        qnoised NSR_with_RP srm.p2.o nsr=True
        # Output just the shot noise limited sensitivity
        qshot NSR_without_RP srm.p2.o nsr=True
        # We could also display the quantum noise and the signal
        # separately by uncommenting these two lines.
        # qnoised noise srm.p2.o
        # pd1 signal srm.p2.o f=fsig
        """
    )
    out = kat.run("xaxis(darmx.f, log, 5, 5k, 100)")
    out.plot(log=True, separate=False)


def ex_10_shifted_beam():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
    import numpy as np

    import finesse

    finesse.configure(plotting=True)
    power = 1.5  # default laser power in Watt
    w0 = 10e-3  # default gaussian waist size in meter3
    alpha = 30  # angle of incidence for both BS in degrees0
    xbeta = 1e-5  # default tilt of both beam splitters in radians
    s1 = 1000  # distance till 1st beamsplitter
    s2 = 400  # distance between beamsplitters
    s3 = 600  # default distance between 2nd beamsplitter and detector
    z0 = -1200  # default waist position, 1200 meter right of laser
    maxtem = 7  # default maxtem
    basescript = f"""
    laser l1 P={power}
    gauss g1 l1.p1.o w0={w0} z={z0}
    modes(maxtem={maxtem})
    space s1 l1.p1 bs1.p1 L={s1}
    beamsplitter bs1 R=1 T=0 alpha={alpha} xbeta={xbeta}
    space s2 bs1.p2 bs2.p1 L={s2}
    beamsplitter bs2 R=1 T=0 alpha=bs1.alpha xbeta=bs1.xbeta
    space s3 bs2.p2 n1.p1 L={s3}
    # Add a 'nothing' component at the location of the detector
    nothing n1
    """
    basekat = finesse.Model()
    basekat.parse(basescript)
    # Simulations: measure beam 600 and 1000 meter after second BS
    s3a = 600
    s3b = 1000
    kat1 = finesse.Model()
    kat1.parse(basescript)
    kat1.parse(
        """
    # Set a 1-D CCD at n1, measure around peak at 0.8 at 600 and 1000m
    ccdline ccd1 node=n1.p1.i xlim=[0.5,1.0] npts=200
    # Also put a full 2-D CCD at the same position
    ccd ccd2 node=n1.p1.i xlim=[-3,3] ylim=[-3,3] npts=200
    """
    )
    out1 = kat1.run(
        f"""
        series(
            noxaxis(name="S600"),
            change(s3.L={s3b}),
            noxaxis(name="S1000")
        )
    """
    )
    # Plot the results
    f, ax = plt.subplots(ncols=2, figsize=(12, 5))
    for i, name, s3val in ([0, "S600", s3a], [1, "S1000", s3b]):
        ax[i].plot(kat1.ccd1.xdata, out1[name]["ccd1"], "r")
        ax[i].set_title(f"{s3val:.0f} meter")
        ax[i].set_xlabel("x/w0")
        ax[i].set_ylabel("intensity")
    f, ax = plt.subplots(ncols=1, figsize=(5, 5))
    pxy_extent = (
        kat1.ccd2.xdata.min(),
        kat1.ccd2.xdata.max(),
        kat1.ccd2.ydata.min(),
        kat1.ccd2.ydata.max(),
    )
    ax.imshow(out1["S600"]["ccd2"].T, aspect="auto", extent=pxy_extent)
    ax.set_xlabel("x/w0")
    ax.set_ylabel("y/w0")
    # Plotting range for x: [peak-dx...peak+dx]
    dx = 0.8
    # Calculate expected Delta and delta phi
    Delta = s2 * np.sin(2 * xbeta) / w0
    dphi = 2 * s2 * np.sin(xbeta) ** 2 / basekat.lambda0 * 360
    kat2 = finesse.Model()
    kat2.parse(basescript)
    kat2.parse(
        f"""
    fline fl1 node=n1.p1.i xlim=[{Delta}-{dx},{Delta}+{dx}] npts=200
    fline fl2 node=n1.p1.i xlim=[-{dx},{dx}] npts=40
    """
    )
    out2 = kat2.run(
        """
        series(
            noxaxis(name='full'),
            change(bs1.xbeta=0),
            noxaxis(name='ref')
        )
    """
    )
    f, ax = plt.subplots(ncols=1, figsize=(8, 5))
    ax.plot(
        kat2.fl1.xdata, np.angle(out2["full"]["fl1"], deg=True), "r", label="measured"
    )
    ax.plot(
        kat2.fl2.xdata + Delta,
        np.angle(out2["ref"]["fl2"], deg=True) + dphi,
        "bp",
        label="prediction",
    )
    ax.set_title("xbeta=1e-5 radian, with predicted peak")
    ax.set_xlabel("x/w0")
    ax.set_ylabel("phase (°)")
    ax.legend(loc="lower center")
    kat3 = finesse.Model()
    kat3.parse(basescript)
    # Note: important to specify n,m
    kat3.parse("amplitude_detector ad1 node=n1.p1.i f=0 n=0 m=0")
    out3 = kat3.run("xaxis(bs1.xbeta, lin, -1.5e-5, 1.5e-5, 40)")
    f, ax = plt.subplots(ncols=1, figsize=(8, 5))
    ax.plot(out3.x[0], np.angle(out3["ad1"], deg=True), "r", label="measured")
    ax.plot(
        out3.x[0],
        2 * s2 * np.sin(out3.x[0]) ** 2 / basekat.lambda0 * 360,
        "bp",
        label="2s·sin²(β)/λ·360°",
    )
    ax.set_title("measured versus calculated")
    ax.set_xlabel("xbeta (radian)")
    ax.set_ylabel("phase (°)")
    ax.legend(loc="upper center")


def ex_angular_radiation_pressure():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import cmath

    import numpy as np

    import finesse
    import finesse.components as fc
    import finesse.detectors as det
    from finesse.analysis.actions import FrequencyResponse, Noxaxis

    finesse.init_plotting()
    L = 3994.5  # [m]
    I = 0.757  # [kg.m^2]
    c = 299792458  # [m.s^-1]
    f_sus = 0.6  # [Hz]
    model = finesse.Model()
    model.fsig.f = 1  # set some initial signal frequency
    model.modes(maxtem=1)  # first order modes for modelling alignment signals
    LASER = model.add(fc.Laser("LASER", P=1000))
    # Add two mirrors for the cavity and attach a pendulum mechanics
    M1 = model.add(fc.Mirror("M1", R=0.986, T=0.014, Rc=1934))
    model.add(
        fc.mechanical.Pendulum(
            "M1_sus", model.M1.mech, mass=np.inf, I_yaw=np.inf, I_pitch=I, fpitch=f_sus
        )
    )
    M2 = model.add(fc.Mirror("M2", R=1, T=0, Rc=2245))
    model.add(
        fc.mechanical.Pendulum(
            "M2_sus", model.M2.mech, mass=np.inf, I_yaw=np.inf, I_pitch=I, fpitch=f_sus
        )
    )
    model.connect(M1.p1, M2.p1, L=L)
    model.connect(LASER.p1, M1.p2)
    model.add(fc.Cavity("cavARM", M2.p1.o))
    model.add(det.PowerDetector("P", M1.p1.o))
    # cavity power
    # Now we compute the decomposition of HARD and SOFT modes into motions of M1 and M2
    g_1 = 1 - float(L / np.abs(M1.Rcx.value))  # input mirror g-factor
    g_2 = 1 - float(L / np.abs(M2.Rcx.value))  # end mirror g-factor
    r = model.add_parameter(
        "r",
        2 / ((g_1 - g_2) + np.sqrt((g_2 - g_1) ** 2 + 4)),
        description="r term in eq.3.5 in T0900511",
    )
    HARD = model.add(
        fc.DegreeOfFreedom("HARD", M1.dofs.F_pitch, -1, M2.dofs.F_pitch, +r.ref)
    )
    SOFT = model.add(
        fc.DegreeOfFreedom("SOFT", M1.dofs.F_pitch, +r.ref, M2.dofs.F_pitch, +1)
    )
    print(HARD.amplitudes)
    print(SOFT.amplitudes)
    print("model.r = ", model.r.eval())
    print("Current HARD amplutudes:", HARD.amplitudes.astype(float))
    print(repr(model.M1.dofs.F_pitch.AC_IN))
    print(repr(model.M1.dofs.F_pitch.AC_OUT))
    print(repr(model.M1.dofs.F_pitch.DC))
    LASER.P = 1410 * 3 / 2  # get to roughly 600kW
    freq_resp = FrequencyResponse(
        np.geomspace(0.1, 10, 2000),
        [HARD.AC.i, SOFT.AC.i],
        [HARD.AC.o, SOFT.AC.o],
    )
    sol = model.run(freq_resp)
    # Alsp compute a single data point of the detectors to compute the power
    # circulating in the cavity
    out = model.run(Noxaxis())
    print(sol.out.shape)
    omega_0 = 2 * np.pi * f_sus
    P = out["P"]
    # Eq 2 from https://opg.optica.org/ao/fulltext.cfm?uri=ao-49-18-3474
    omega_plus = np.sqrt(
        omega_0**2
        + P
        * L
        / (I * c)
        * (-(g_1 + g_2) + cmath.sqrt(4 + (g_1 - g_2) ** 2))
        / (1 - g_1 * g_2)
    )
    omega_minus = np.sqrt(
        omega_0**2
        + P
        * L
        / (I * c)
        * (-(g_1 + g_2) - cmath.sqrt(4 + (g_1 - g_2) ** 2))
        / (1 - g_1 * g_2)
    )
    axs = finesse.plotting.bode(sol.f, sol["HARD.AC.o", "HARD.AC.i"], label="HARD")
    axs = finesse.plotting.bode(
        sol.f, sol["SOFT.AC.o", "SOFT.AC.i"], label="SOFT", axs=axs
    )
    axs[0].vlines(
        omega_0 / 2 / np.pi,
        -70,
        60,
        ls="--",
        label=f"$\\omega_{{0}} = {omega_0.real / 2 / np.pi:0.2f}$Hz",
        color="k",
        zorder=-100,
    )
    axs[0].vlines(
        omega_plus.real / 2 / np.pi,
        -70,
        60,
        ls="--",
        label=f"$\\omega_{{+}} = {omega_plus.real / 2 / np.pi:0.2f}$Hz",
        zorder=-100,
    )
    axs[0].vlines(
        omega_minus.real / 2 / np.pi,
        -70,
        60,
        color="red",
        ls="--",
        label=f"$\\omega_{{-}} = {omega_minus.real / 2 / np.pi:0.2f}$Hz",
        zorder=-100,
    )
    axs[0].margins(0)
    axs[0].set_title(
        f"E.Hirose, Appl. Opt. 49, 3474-3484 (2010)\nEq.2 vs FINESSE3, P={P / 1e3:0.0f}kW"
    )
    axs[0].legend()
    axs[0].set_ylabel("Magnitude [dB - rad/Nm]")
    LASER.P = 100
    sol = model.run(freq_resp)
    axs = finesse.plotting.bode(sol.f, sol["HARD.AC.o", "HARD.AC.i"], label="HARD")
    finesse.plotting.bode(sol.f, sol["SOFT.AC.o", "SOFT.AC.i"], label="SOFT", axs=axs)
    LASER.P = 1410 * 3 / 2  # get to roughly 600kW
    model.r = 1
    # Or alternatively you can replace the amplitudes directly with something like
    # HARD.amplitudes[:] = [-1, +1]
    # SOFT.amplitudes[:] = [+1, +1]
    sol = model.run(freq_resp)
    axs = finesse.plotting.bode(sol.f, sol["HARD.AC.o", "HARD.AC.i"], label="HARD")
    finesse.plotting.bode(sol.f, sol["SOFT.AC.o", "SOFT.AC.i"], label="SOFT", axs=axs)


def ex_cavity_eigenmodes():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
    import numpy as np

    import finesse
    from finesse.knm import Map
    from finesse.utilities.maps import circular_aperture

    finesse.init_plotting()
    model = finesse.Model()
    model.parse(
        """
        l l1
        mod mod1 f=9.1M midx=0.1
        m m1 R=0.984 T=0.014 Rc=-1940
        m m2 R=1 T=0 Rc=2245
        link(l1, mod1, m1, 3994, m2)
        cav cavity m2.p1.o
        modes(maxtem=2)
        fd E_arm m2.p1.i l1.f  # Field detector
        """
    )
    sol = model.run(
        """
        series(
            eigenmodes(cavity, -mod1.f, name="l9"),
            eigenmodes(cavity, 0,       name="c0"),
            eigenmodes(cavity, mod1.f,  name="u9"),
            xaxis(m2.phi, lin, -10, 190, 500, name="scan")
        )
        """
    )
    print(sol)
    sol["c0"]
    plt.semilogy(abs(sol["scan"]["E_arm"]) ** 2)
    sol["c0"].eigvectors
    print(abs(sol["c0"].eigvalues))
    print(np.angle(sol["c0"].eigvalues))
    from finesse.plotting import plot_field

    plot_field(model.homs, sol["c0"].eigvectors[:, 0], model.cavity.source.q)
    sol["c0"].plot_field(mode_idx=1)
    x = y = np.linspace(-0.17, 0.17, 100)
    ligo_ap = circular_aperture(x, y, 0.17)
    ligo_ap[:, abs(y) >= 0.163] = 0
    model.m1.surface_map = Map(x, y, amplitude=ligo_ap)
    model.m2.surface_map = Map(x, y, amplitude=ligo_ap)
    # Plot the aperture we are applying.
    plt.pcolormesh(x, y, ligo_ap, rasterized=True, shading="auto")
    plt.gca().set_aspect("equal")
    plt.colorbar()
    sol_aperture = model.run("eigenmodes(cavity, 0, name='c0')")
    for i in range(len(model.homs)):
        plot_field(model.homs, sol_aperture.eigvectors[:, i], model.cavity.source.q)


def ex_coupled_cavity_commands():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
    import numpy as np

    import finesse

    finesse.init_plotting()
    model = finesse.script.parse(
        """
    ###############################################################################
    ###   length definitions
    ###############################################################################
    variable Larm 3994.47
    variable LPR23 16.164  # distance between PR2 and PR3
    variable LPR3BS 19.538 # distance between PR3 and BS
    variable lmich 5.342   # average length of MICH
    variable lschnupp 0.08
    variable lPRC (3+0.5)*c0/(2*f1) # T1000298 Eq2.1, N=3
    variable f1 9099471
    variable f2 5*f1
    variable nsilica 1.45
    variable Mloss 0
    laser L0 P=1
    mod mod1 f=f1 midx=0.2
    link(L0, mod1, PRM)
    m PRM T=0.03 L=0 Rc=11.009 phi=0
    s lp1 PRM.p2 PR2.p1 L=lPRC-LPR3BS-LPR23-lmich
    bs PR2 T=250u L=Mloss alpha=-0.79 Rc=-4.545
    s lp2 PR2.p2 PR3.p1 L=LPR23
    bs PR3 T=0 L=Mloss alpha=0.615 Rc=36.027
    s lp3 PR3.p2 BS.p1 L=LPR3BS
    ###############################################################################
    ###   BS
    ###############################################################################
    bs BS R=1 L=0 alpha=45
    ###############################################################################
    ###   Yarm
    ###############################################################################
    # Distance from beam splitter to Y arm input mirror
    s ly1 BS.p2 ITMYlens.p1 L=lmich-lschnupp/2-ITMYsub.L*ITMYsub.nr
    lens ITMYlens f=34000
    s ly2 ITMYlens.p2 ITMYAR.p1
    m ITMYAR R=0 L=20u
    s ITMYsub ITMYAR.p2 ITMY.p1 L=0.2 nr=nsilica
    m ITMY T=0.014 L=Mloss Rc=-1934
    s LY ITMY.p2 ETMY.p1 L=Larm
    m ETMY T=5u L=Mloss Rc=2245
    pd Parm ETMY.p1.i
    pd Prc PRM.p2.o
    modes(even, maxtem=12)
    ad pu9 PRM.p2.i f=+f1
    ad pl9 PRM.p2.i f=-f1
    mathd prc_cost log10(Parm/280)+log10(abs(pu9))+log10(abs(pl9))
    """
    )
    # make a copy so that we can add different cav commands
    _model = model.deepcopy()
    _model.parse("cav cavYARM ETMY.p1.o")
    # nodes that will be defined by the ITM and ETM geometry
    print(_model.cavYARM.path.nodes)
    _model = model.deepcopy()
    _model.parse("cav cavYARM ETMY.p1.o")
    # Maximize the arm power with respect to the PRM curvature and ITMY lens
    # There are multiple optimisations that could be done here, depends on what
    # parameters you want to vary. This model is naturally astigmatic though due
    # to the off-axis telescope.
    opt_sol = _model.run(
        "maximize(Parm, [ITMYlens.f, PRM.Rcx, PRM.Rcy, PRM.phi, ETMY.phi])"
    )
    print("HOMs Parm =       ", _model.run("noxaxis()")["Parm"])
    _model.modes("off")
    print("Plane-wave Parm = ", _model.run("noxaxis()")["Parm"])
    # the plane-wave and HOM powers should be nearly identical, a sign that there
    # is no mode-mismatch. If so, we then update the original model to use these
    # new values
    model.ITMYlens.f = opt_sol.x[0]
    model.PRM.Rcx = opt_sol.x[1]
    model.PRM.Rcy = opt_sol.x[2]
    # Given the model is now in a perfectly matched state we can also take the
    # beam parameter at the input of the PRM and set the original model to use that
    model.PRM.p1.i.q = _model.PRM.p1.i.q
    model.gausses
    model.ITMYlens.f = np.inf  # opt_sol.x[0] # use opt_sol for perfect matching
    model.mismatches_table()

    def run(model, orders):
        model.modes("off")
        plane_wave = model.run()
        homs = []
        for n in orders:
            model.modes("even", maxtem=n)
            with model.temporary_parameters():
                homs.append(
                    model.run(
                        """
                    series(
                        maximize(Parm, [PRM.phi, ETMY.phi]),
                        noxaxis() # get detector outputs
                    )
                """
                    )["noxaxis"]
                )
        return plane_wave, homs

    orders = range(0, 10, 2)
    _model = model.deepcopy()
    _model.parse("cav cavYARM ETMY.p1.o")
    plane_wave, homs = run(_model, orders)
    _model = model.deepcopy()
    _model.parse("cav cavPRC PRM.p2.o via=ITMY.p1.i")
    plane_wave2, homs2 = run(_model, orders)
    _model = model.deepcopy()
    _model.parse("cav cavYARM ETMY.p1.o")
    _model.parse("cav cavPRC PRM.p2.o via=ITMY.p1.i")
    plane_wave3, homs3 = run(_model, orders)
    _model = model.deepcopy()
    _model.parse("cav cavYARM ETMY.p1.o via=PRM.p2.i")
    plane_wave4, homs4 = run(_model, orders)
    plt.plot(orders, tuple(_["Parm"] for _ in homs), label="cav ARM")
    plt.plot(orders, tuple(_["Parm"] for _ in homs2), label="cav PRC")
    plt.plot(orders, tuple(_["Parm"] for _ in homs3), label="cav ARM+PRC")
    plt.plot(orders, tuple(_["Parm"] for _ in homs4), label="long cav", ls="--")
    plt.hlines(plane_wave["Parm"], 0, 8, ls=":", color="m", label="plane-wave")
    plt.legend()
    plt.xlabel("maxtem")
    plt.ylabel("P arm")
    plt.title(repr(model.ITMYlens.f))


def ex_frequency_dependant_squeezing():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import matplotlib.pyplot as plt
    import numpy as np

    import finesse

    finesse.init_plotting()
    base = finesse.Model()
    base.parse(
        """
    variable ITM_T 0.014
    variable L_ARM 40k
    variable FC_detune 7.3 # Hz
    l l1 P=165
    link(l1, PRM)
    # ------------------------------------------------------------------------------
    # Power recycling cavity
    # ------------------------------------------------------------------------------
    m PRM T=0.03 L=40u phi=0
    s PRC PRM.p2 BS.p1 L=53.0
    bs BS R=0.5 T=0.5
    # ------------------------------------------------------------------------------
    # X-arm
    # ------------------------------------------------------------------------------
    s lx BS.p3 ITMX.p1 L=0
    m ITMX T=ITM_T L=0u phi=0
    s LX ITMX.p2 ETMX.p1 L=L_ARM
    m ETMX T=5u L=40u phi=0
    # ------------------------------------------------------------------------------
    # Y-arm
    # ------------------------------------------------------------------------------
    s ly BS.p2 ITMY.p1 L=0
    m ITMY T=ITM_T L=0u phi=90
    s LY ITMY.p2 ETMY.p1 L=L_ARM
    m ETMY T=5u L=40u phi=90
    # Signal recycling cavity
    s SRC BS.p4 SRM.p1 L=20
    m SRM T=0.02 L=0 phi=0
    # ------------------------------------------------------------------------------
    # Output optics
    # ------------------------------------------------------------------------------
    dbs OFI
    link(SRM.p2, OFI.p1)
    readout_dc AS OFI.p3.o
    # ------------------------------------------------------------------------------
    # suspensions
    # ------------------------------------------------------------------------------
    free_mass ITMX_sus ITMX mass=320
    free_mass ETMX_sus ETMX mass=320
    free_mass ITMY_sus ITMY mass=320
    free_mass ETMY_sus ETMY mass=320
    # ------------------------------------------------------------------------------
    # Degrees of Freedom
    # ------------------------------------------------------------------------------
    dof STRAIN LX.dofs.h +1  LY.dofs.h -1
    # signal generator
    sgen sig STRAIN
    qnoised NSR AS.p1.i nsr=True
    fsig(1)
    """
    )
    model = base.deepcopy()
    model.parse(
        """
    sq sqz db=10
    link(sqz, OFI.p2)
    """
    )
    with model.temporary_parameters():
        # No squeezing
        model.sqz.db = 0
        model.sqz.angle = 0
        sol_no_sqz = model.run("xaxis(fsig, log, 1, 5k, 100)")
        # Squeezing
        model.sqz.db = 10
        model.sqz.angle = 90
        sol_shot_sqz = model.run("xaxis(fsig, log, 1, 5k, 100)")
        # Squeezing
        model.sqz.db = 10
        model.sqz.angle = 0
        sol_rp_sqz = model.run("xaxis(fsig, log, 1, 5k, 100)")
    plt.loglog(sol_no_sqz.x1, abs(sol_no_sqz["NSR"]), label="No squeezing")
    plt.loglog(sol_shot_sqz.x1, abs(sol_shot_sqz["NSR"]), label="Shot-noise squeezing")
    plt.loglog(sol_rp_sqz.x1, abs(sol_rp_sqz["NSR"]), label="Rad. pressure squeezing")
    plt.legend()
    plt.xlabel("Frequency [Hz]")
    plt.ylabel("Sensitivity [h/sqrt{Hz}]")
    with model.temporary_parameters():
        model.FC_detune.value = 7.2
        model.sqz.db = 10
        # Give the variable references shorter names
        fsig = model.fsig.f.ref
        delta_f = model.FC_detune.ref
        # set the squeezing angle to be a symbolic function
        model.sqz.angle = 45 + 180 / np.pi * np.arctan(
            (fsig**2 - delta_f**2) / (fsig**2 + delta_f**2)
        )
        print(model.sqz.angle)  # print to see how it turned out
        sol_fc_sqz = model.run("xaxis(fsig, log, 1, 5k, 100)")
    plt.loglog(sol_no_sqz.x1, abs(sol_no_sqz["NSR"]), label="No squeezing")
    plt.loglog(sol_shot_sqz.x1, abs(sol_shot_sqz["NSR"]), label="Shot-noise squeezing")
    plt.loglog(sol_rp_sqz.x1, abs(sol_rp_sqz["NSR"]), label="Rad. pressure squeezing")
    plt.loglog(sol_fc_sqz.x1, abs(sol_fc_sqz["NSR"]), label="Filter cavity")
    plt.legend()
    plt.xlabel("Frequency [Hz]")
    plt.ylabel("Sensitivity [h/sqrt{Hz}]")
    # Run an optimiser to find ideal rotation
    sol_opt_sqz = model.run(
        "xaxis(fsig, log, 1, 5k, 100, pre_step=minimize(NSR, sqz.angle))"
    )
    plt.loglog(sol_no_sqz.x1, abs(sol_no_sqz["NSR"]), label="No squeezing")
    plt.loglog(sol_fc_sqz.x1, abs(sol_fc_sqz["NSR"]), label="Filter cavity (symbolic)")
    plt.loglog(
        sol_opt_sqz.x1,
        abs(sol_opt_sqz["NSR"]),
        label="Filter cavity (optimized)",
        ls="--",
    )
    plt.legend()
    plt.xlabel("Frequency [Hz]")
    plt.ylabel("Sensitivity [h/sqrt{Hz}]")
    # Extract every optimisisation solution (`x`) from all the pre-step minimize actions
    opt_angle = sol_opt_sqz["pre_step", "minimize"].x
    plt.semilogx(sol_opt_sqz.x1, opt_angle)
    plt.xlabel("Frequency [Hz]")
    plt.ylabel("Optimal squeezing angle [deg]")
    with model.temporary_parameters():
        model.FC_detune.value = 7.2
        model.SRM.phi += 0.1  # Detune slightly
        model.sqz.db = 10
        # Give the variable references shorter names
        fsig = model.fsig.f.ref
        delta_f = model.FC_detune.ref
        # set the squeezing angle to be a symbolic function
        model.sqz.angle = 45 + 180 / np.pi * np.arctan(
            (fsig**2 - delta_f**2) / (fsig**2 + delta_f**2)
        )
        sol_fc_sqz = model.run("xaxis(fsig, log, 1, 5k, 100)")
        # Run an optimiser to find ideal rotation
        model.sqz.angle = 0
        sol_opt_sqz = model.run(
            "xaxis(fsig, log, 1, 5k, 100, pre_step=minimize(NSR, sqz.angle))"
        )
        plt.loglog(sol_no_sqz.x1, abs(sol_no_sqz["NSR"]), label="No squeezing")
        plt.loglog(
            sol_fc_sqz.x1, abs(sol_fc_sqz["NSR"]), label="Filter cavity (symbolic)"
        )
        plt.loglog(
            sol_opt_sqz.x1, abs(sol_opt_sqz["NSR"]), label="Filter cavity (optimized)"
        )
        plt.legend()
        plt.xlabel("Frequency [Hz]")
        plt.ylabel("Sensitivity [h/sqrt{Hz}]")
        plt.title("Detuned SRC")
        plt.figure()
        # Extract every optimisisation solution (`x`) from all the pre-step minimize actions
        opt_angle = sol_opt_sqz["pre_step", "minimize"].x
        plt.semilogx(sol_opt_sqz.x1, opt_angle)
        plt.xlabel("Frequency [Hz]")
        plt.ylabel("Squeezing angle [deg]")
        plt.title("Optimised squeezing angle")
    model = base.deepcopy()  # make a copy so we can add extra features
    model.parse(
        """
    sq sqz db=10
    dbs FI
    m FC1 L=0 T=0.0015
    s sFC FC1.p2 FC2.p1 L=4e3
    m FC2 L=0 T=0 phi=-0.022
    link(sqz, FI.p1)
    link(FI.p3, FC1)
    link(FI.p4, OFI.p2)
    """
    )
    model.sqz.db = 10
    model.sqz.angle = 90
    model.FC1.T = 0.0015
    model.FC1.R = (
        1 - model.FC1.T.ref
    )  # Make sure we update the R correctly when we change T
    model.FC2.phi = -0.022
    # Optimise the NSR at the corner frequency
    model.fsig.f = 7.2
    # Run the optimisation a few times to reach a better level
    for _ in range(3):
        model.run("minimize(NSR, [FC1.T, FC2.phi, sFC.L])")
    # finally compute the NSR
    sol_fc2_sqz = model.run("xaxis(fsig, log, 1, 5k, 100)")
    plt.loglog(sol_no_sqz.x1, abs(sol_no_sqz["NSR"]), label="No squeezing")
    plt.loglog(sol_fc2_sqz.x1, abs(sol_fc2_sqz["NSR"]), label="Filter cavity model")
    plt.loglog(
        sol_opt_sqz.x1, abs(sol_opt_sqz["NSR"]), label="Filter cavity (ideal)", ls="--"
    )
    plt.legend()
    plt.xlabel("Frequency [Hz]")
    plt.ylabel("Sensitivity [h/sqrt{Hz}]")
    print("FC1.T =", model.FC1.T)
    print("FC2.phi =", model.FC2.phi)
    print("FC L =", model.sFC.L)


def ex_lock_actions():
    import matplotlib

    # Suppress plot showing
    matplotlib.use("Agg")
    import finesse
    from finesse.analysis.actions import (
        CheckLinearity,
        DragLocks,
        GetErrorSignals,
        OptimiseRFReadoutPhaseDC,
        RunLocks,
        SensingMatrixDC,
    )
    from finesse.analysis.actions.locks import SetLockGains

    finesse.configure(plotting=True)
    kat = finesse.Model()
    kat.parse(
        """
        # Adding a laser and modulator for use in
        # Pound-Drever-Hall-like signals.
        l L0 P=1
        s l_mod1 L0.p1 eo1.p1
        mod eo1 10M 0.1
        s s0 eo1.p2 BS.p1
        bs BS R=0.5 T=0.5
        # North Arm
        s s1 BS.p2 NI.p1
        m NI R=0.99 T=0.01 Rc=1429 phi=90
        s CAV NI.p2 NE.p1 L=10
        m NE R=0.991 T=0.009 Rc=1430 phi=90
        #East Arm
        s s2 BS.p3 EI.p1
        m EI R=0.99 T=0.01 Rc=1429 phi=0
        s CAV2 EI.p2 EE.p1 L=10
        m EE R=0.991 T=0.009 Rc=1430 phi=0
        dof NEz NE.dofs.z +1
        dof EEz EE.dofs.z +1
        dof NIz NI.dofs.z +1
        readout_rf rd_pdh1 NI.p1.o f=10M
        readout_rf rd_pdh2 EI.p1.o f=10M
        readout_rf rd_DF BS.p4.o f=10M
        # Two locks to put the cavities in resonance, and
        # one to move to the dark fringe.
        lock cav1_lock rd_pdh1.outputs.I NEz.DC 1 1e-9
        lock cav2_lock rd_pdh2.outputs.I EEz.DC 1 1e-9
        lock DF_lock rd_DF.outputs.I NIz.DC 1 1e-9
        cav cav1 NI.p2.o
        cav cav2 EI.p2.o
        modes(maxtem=5)
        """
    )
    kat.NEz.DC += 0.01
    kat.EEz.DC += 0.001
    sol_errs = kat.run(GetErrorSignals("cav1_lock", "DF_lock"))
    print("Initial Error Signals:", sol_errs.results, end="\n\n")
    sol_phases = kat.run(OptimiseRFReadoutPhaseDC("NEz", "rd_pdh1", "EEz", "rd_pdh2"))
    print("Optimal Readout Phases:", sol_phases.phases, end="\n\n")
    kat.run(SetLockGains(optimize_phase=True))
    sol_matrix = kat.run(
        SensingMatrixDC(["NEz", "EEz", "NIz"], ["rd_pdh1", "rd_pdh2", "rd_DF"])
    )
    print("Sensing Matrix:", sol_matrix, end="\n\n", sep="\n")
    kat.run(CheckLinearity(plot_results=True))
    kat.run(
        RunLocks(
            method="newton",
            display_progress=False,
            max_iterations=100,
            optimize_phase=True,
            scale_factor=1,
        )
    )
    print("Results of dragging locks:")
    kat.run(
        DragLocks(
            method="newton",
            parameters=["NE.Rcx"],
            stop_points=[1],
            relative=True,
            max_iterations=100,
        )
    )



import finesse
import numpy as np
from finesse.gaussian import optimise_HG00_q_scipy
import pytest
from scipy.optimize import OptimizeResult
from finesse.analysis.actions import DCFields
from finesse.components import Mirror, Laser
from finesse.knm.tools import make_bayerhelms_matrix


@pytest.mark.parametrize("z", [0, -1, 2])
@pytest.mark.parametrize("w0", [0.011, 0.006, 0.003])
def test_optimise_HG00_q_scipy(z, w0, model):
    qi = finesse.BeamParam(w0=6e-3, z=0)
    qo = finesse.BeamParam(w0=w0, z=z)
    model.add(Laser("l1"))
    model.add(Mirror("mirror", T=1, R=0))
    model.connect(model.l1.p1, model.mirror.p1)
    model.mirror.p1.i.q = qi
    model.mirror.p2.o.q = qo
    model.modes("even", maxtem=20)

    sol = model.run(DCFields())
    Eo = sol["mirror.p2.o"].squeeze()
    result = optimise_HG00_q_scipy(
        Eo,
        qo,
        model.homs,
        accuracy=1e-9,
        method="nelder-mead",
    )
    result2 = optimise_HG00_q_scipy(
        Eo,
        qo,
        model.homs,
        accuracy=1e-9,
        full_output=True,
        method="nelder-mead",
    )

    kmat = make_bayerhelms_matrix(
        qo,
        result2[0].q,
        qo,
        result2[1].q,
        0,
        0,
        select=model.homs,
        reverse_gouy=False,
    )
    field = kmat.data @ Eo

    assert np.allclose(model.mirror.p1.i.qx.q, result[0].q, rtol=1e-3)
    assert np.allclose(model.mirror.p1.i.qy.q, result[1].q, rtol=1e-3)
    assert np.allclose(result2[-1]["field"], field)
    assert isinstance(result2[-1]["res"], OptimizeResult)


@pytest.mark.parametrize("z", [0, -1, 2])
@pytest.mark.parametrize("w0", [0.011, 0.006, 0.003])
def test_optimal_q_detector(z, w0):
    model = finesse.script.parse(
        f"""
    l l1
    mirror n1 T=1 R=0
    link(l1, n1)
    gauss g1 n1.p1.i w0=6m z=0
    gauss g2 n1.p2.o w0={w0} z={z}
    fd Ei n1.p1.i f=0
    fd Eo n1.p2.o f=0
    modes(even, maxtem=20)
    optimal_q_detector Q n1.p2.o 0
    """
    )
    sol = model.run()

    assert np.allclose(model.n1.p1.i.qx.q, sol["Q"][0].q, rtol=1e-3)
    assert np.allclose(model.n1.p1.i.qy.q, sol["Q"][1].q, rtol=1e-3)



import pytest
from finesse.components.laser import Laser
from finesse.script import KATSPEC, unparse


@pytest.fixture
def laser_model(model):
    model.add(Laser("l1"))
    return model


@pytest.mark.parametrize(
    "kwargs,directive_defaults,argument_defaults,prefer_keywords,expected",
    (
        ## No TEM commands.
        (None, False, False, False, ""),
        (None, False, False, True, ""),
        (None, False, True, False, ""),
        (None, False, True, True, ""),
        (None, True, False, False, "tem(l1, 0, 0, 1.0)"),
        (None, True, False, True, "tem(laser=l1, n=0, m=0, factor=1.0)"),
        (None, True, True, False, "tem(l1, 0, 0, 1.0, 0.0)"),
        (None, True, True, True, "tem(laser=l1, n=0, m=0, factor=1.0, phase=0.0)"),
        ## Non-empty commands.
        # No phase.
        ({"n": 0, "m": 0, "factor": 0.5}, False, False, False, "tem(l1, 0, 0, 0.5)"),
        (
            {"n": 0, "m": 0, "factor": 0.5},
            False,
            False,
            True,
            "tem(laser=l1, n=0, m=0, factor=0.5)",
        ),
        (
            {"n": 0, "m": 0, "factor": 0.5},
            False,
            True,
            False,
            "tem(l1, 0, 0, 0.5, 0.0)",
        ),
        (
            {"n": 0, "m": 0, "factor": 0.5},
            False,
            True,
            True,
            "tem(laser=l1, n=0, m=0, factor=0.5, phase=0.0)",
        ),
        # Phase set to default.
        (
            {"n": 0, "m": 0, "factor": 0.5, "phase": 0.0},
            False,
            False,
            False,
            "tem(l1, 0, 0, 0.5)",
        ),
        (
            {"n": 0, "m": 0, "factor": 0.5, "phase": 0.0},
            False,
            False,
            True,
            "tem(laser=l1, n=0, m=0, factor=0.5)",
        ),
        (
            {"n": 0, "m": 0, "factor": 0.5, "phase": 0.0},
            False,
            True,
            False,
            "tem(l1, 0, 0, 0.5, 0.0)",
        ),
        (
            {"n": 0, "m": 0, "factor": 0.5, "phase": 0.0},
            False,
            True,
            True,
            "tem(laser=l1, n=0, m=0, factor=0.5, phase=0.0)",
        ),
        # Phase not set to default.
        (
            {"n": 0, "m": 0, "factor": 0.5, "phase": 45},
            False,
            False,
            False,
            "tem(l1, 0, 0, 0.5, 45.0)",
        ),
        (
            {"n": 0, "m": 0, "factor": 0.5, "phase": 45},
            False,
            False,
            True,
            "tem(laser=l1, n=0, m=0, factor=0.5, phase=45.0)",
        ),
        (
            {"n": 0, "m": 0, "factor": 0.5, "phase": 45.0},
            False,
            True,
            False,
            "tem(l1, 0, 0, 0.5, 45.0)",
        ),
        (
            {"n": 0, "m": 0, "factor": 0.5, "phase": 45.0},
            False,
            True,
            True,
            "tem(laser=l1, n=0, m=0, factor=0.5, phase=45.0)",
        ),
    ),
)
def test_tem(
    laser_model,
    kwargs,
    directive_defaults,
    argument_defaults,
    prefer_keywords,
    expected,
):
    adapter = KATSPEC.commands["tem"]

    if kwargs:
        laser_model.l1.tem(**kwargs)

    dump = next(iter(adapter.getter(adapter, laser_model)))
    script = unparse(
        dump,
        directive_defaults=directive_defaults,
        argument_defaults=argument_defaults,
        prefer_keywords=prefer_keywords,
    )
    assert script == expected



"""Test cases for shallow copying models."""

import copy

import numpy as np
from networkx.utils import graphs_equal
import pytest


@pytest.fixture(scope="module")
def shallow_copied_model(complicated_ifo_for_copying):
    """Shallow copied complicated model.

    Tests that use this fixture rely on all of the checked parameters NOT BEING EMPTY.
    That means the model here should have components, cavities and detectors.
    """
    if (
        not complicated_ifo_for_copying.components
        or not complicated_ifo_for_copying.cavities
        or not complicated_ifo_for_copying.detectors
    ):
        raise ValueError(
            "shallow_copied_model requires a model with components, cavities and detectors"
        )

    return complicated_ifo_for_copying, copy.copy(complicated_ifo_for_copying)


def test_shallow_copied_model_ids_dont_match(shallow_copied_model):
    """Check that shallow copied model IDs are different."""
    original, copied = shallow_copied_model
    assert id(original) != id(copied)


def test_deep_copied_model_have_same_properties(shallow_copied_model):
    """Check that shallow copied models have the same properties."""
    original, copied = shallow_copied_model
    assert np.all(original.homs == copied.homs)
    assert original._settings.fsig == copied._settings.fsig
    assert original._settings.EPSILON0_C == copied._settings.EPSILON0_C
    assert original._settings.UNIT_VACUUM == copied._settings.UNIT_VACUUM
    assert original._settings.lambda0 == copied._settings.lambda0
    assert original._settings.f0 == copied._settings.f0
    assert original._settings.k0 == copied._settings.k0
    assert original._settings.is_modal == copied._settings.is_modal
    assert original._settings.num_HOMs == copied._settings.num_HOMs
    assert original._settings.max_n == copied._settings.max_n
    assert original._settings.max_m == copied._settings.max_m
    assert original._settings.x_scale == copied._settings.x_scale
    assert original.lambda0 == copied.lambda0


def test_shallow_copied_model_networks_match(shallow_copied_model):
    """Check that shallow copied model networks are the same."""
    original, copied = shallow_copied_model
    assert original.network == copied.network


def test_shallow_copied_model_optical_networks_match(shallow_copied_model):
    """Check that shallow copied model optical networks are the same."""
    original, copied = shallow_copied_model
    assert graphs_equal(original.optical_network, copied.optical_network)


def test_shallow_copied_model_cavities_match(shallow_copied_model):
    """Check that shallow copied model cavities are the same."""
    original, copied = shallow_copied_model
    assert original.cavities == copied.cavities


def test_shallow_copied_model_gauss_commands_match(shallow_copied_model):
    """Check that shallow copied model gauss commands are the same."""
    original, copied = shallow_copied_model
    assert original.gausses == copied.gausses


def test_shallow_copied_model_components_match(shallow_copied_model):
    """Check that shallow copied model components are the same."""
    original, copied = shallow_copied_model
    assert original.components == copied.components


def test_shallow_copied_model_detectors_match(shallow_copied_model):
    """Check that shallow copied model detectors are the same."""
    original, copied = shallow_copied_model
    assert original.detectors == copied.detectors


def test_shallow_copied_model_elements_match(shallow_copied_model):
    """Check that shallow copied model elements are the same."""
    original, copied = shallow_copied_model
    assert original.elements == copied.elements



"""Dielectric interface type components representing physical mirrors."""

import logging
import types
import finesse
import numpy as np

from finesse.parameter import float_parameter, bool_parameter
from finesse.utilities import refractive_index

from finesse.components.general import (
    InteractionType,
    NoiseType,
    LocalDegreeOfFreedom,
)
from finesse.components.surface import Surface
from finesse.components.node import NodeType, NodeDirection
from finesse.tracing import abcd
from finesse.symbols import Constant, Variable, Matrix

LOGGER = logging.getLogger(__name__)


@float_parameter("R", "Reflectivity", validate="_check_R", post_validate="check_rtl")
@float_parameter("T", "Transmission", validate="_check_T", post_validate="check_rtl")
@float_parameter("L", "Loss", validate="_check_L", post_validate="check_rtl")
@float_parameter(
    "phi", "Microscopic tuning (360 degrees = 1 default wavelength)", units="degrees"
)
@float_parameter(
    "Rcx",
    "Radius of curvature (x)",
    units="m",
    validate="_check_Rc",
    is_geometric=True,
)
@float_parameter(
    "Rcy",
    "Radius of curvature (y)",
    units="m",
    validate="_check_Rc",
    is_geometric=True,
)
@float_parameter("xbeta", "Yaw misalignment", units="radians")
@float_parameter("ybeta", "Pitch misalignment", units="radians")
@bool_parameter("misaligned", "Misaligns mirror reflection (R=0 when True)")
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Mirror(Surface):
    """The mirror component represents a thin dielectric surface with associated
    properties such as reflectivity, tuning, and radius of curvature. Mirror components
    are nominally at normal incidence to the beams. It has two optical ports `p1` and
    `p2` which describes the two beams incident on either side of this surface. The
    surface normal points out of the mirror on the p1 side. A mirror also has a
    mechanical port `mech` which has nodes for longitudinal, yaw, and pitch motions.
    These mechanical nodes are purely for exciting small signal oscillations of the
    mirror. Static offsets in longitudinal displacements are set by the `phi` parameter
    (in units of degrees), yaw by the `xbeta` parameter, and pitch the `ybeta`
    parameter.

    Parameters
    ----------
    name : str
        Name of newly created mirror.

    R : float, optional
        Reflectivity of the mirror; defaults to 0.5.

    T : float, optional
        Transmittance of the mirror; defaults to 0.5.

    L : float, optional
        Loss of the mirror; defaults to 0.0.

    phi : float, optional
        Tuning of the mirror (in degrees); defaults to 0.0.

    Rc : float or container of two floats, optional
        The radius of curvature of the mirror (in metres);
        defaults to a flat mirror (`Rc=np.inf`).
        Astigmatic mirrors can also be set with `Rc=(Rcx, Rcy)`.
        A positive value results in a concave mirror on the p1
        side of the mirror.

    xbeta, ybeta : float, optional
        Misalignment of the mirror in yaw and pitch in units of radians

    misaligned : bool, optional
        When True the mirror will be significantly misaligned and
        assumes any reflected beam is dumped. Transmissions will
        still occur.

    imaginary_transmission : bool, optional
        Convention for the transmission and reflection reciprocity relations. 'True'
        uses imaginary transmission and equal real reflection from both sides 'False'
        uses real transmission, negative reflection from side 1, and positive reflection
        from side 2. defaults to True

    Attributes
    ----------
    Attributes are set via the Python API and not available via KatScript.

    surface_map : :class:`finesse.knm.maps.Map`
        Decsribes the surface distortion of this mirror component.
        Coordinate system to the map is right-handed with the postive-z
        direction as the surface normal on the port 1 side of the mirror.
    """

    def __init__(
        self,
        name,
        R=None,
        T=None,
        L=None,
        phi=0,
        Rc=np.inf,
        xbeta=0,
        ybeta=0,
        misaligned=False,
        imaginary_transmission=True,
    ):
        super().__init__(name, R, T, L, phi, Rc, xbeta, ybeta)
        self.misaligned = misaligned
        self.surface_map = None
        self.imaginary_transmission = imaginary_transmission

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p2", NodeType.OPTICAL)
        self.p2._add_node("i", NodeDirection.INPUT)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        # Optic to optic couplings
        self._register_node_coupling(
            "P1i_P1o",
            self.p1.i,
            self.p1.o,
            interaction_type=InteractionType.REFLECTION,
            # enabled_check=lambda: self.R > 0 and not self.R.is_changing
        )
        self._register_node_coupling(
            "P2i_P2o",
            self.p2.i,
            self.p2.o,
            interaction_type=InteractionType.REFLECTION,
            # enabled_check=lambda: self.R > 0 and not self.R.is_changing
        )
        self._register_node_coupling(
            "P1i_P2o",
            self.p1.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
            # enabled_check=lambda: self.T > 0 and not self.T.is_changing
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
            # enabled_check=lambda: self.T > 0 and not self.T.is_changing
        )

        # Mirror motion couplings
        self._add_port("mech", NodeType.MECHANICAL)
        self.mech._add_node("z", NodeDirection.BIDIRECTIONAL)
        self.mech._add_node("yaw", NodeDirection.BIDIRECTIONAL)
        self.mech._add_node("pitch", NodeDirection.BIDIRECTIONAL)
        self.mech._add_node("F_z", NodeDirection.BIDIRECTIONAL)
        self.mech._add_node("F_yaw", NodeDirection.BIDIRECTIONAL)
        self.mech._add_node("F_pitch", NodeDirection.BIDIRECTIONAL)

        # Optic to motion couplings
        self._register_node_coupling("P1i_Fz", self.p1.i, self.mech.F_z)
        self._register_node_coupling("P2i_Fz", self.p2.i, self.mech.F_z)
        self._register_node_coupling("P1o_Fz", self.p1.o, self.mech.F_z)
        self._register_node_coupling("P2o_Fz", self.p2.o, self.mech.F_z)
        self._register_node_coupling("P1i_Fyaw", self.p1.i, self.mech.F_yaw)
        self._register_node_coupling("P2i_Fyaw", self.p2.i, self.mech.F_yaw)
        self._register_node_coupling("P1o_Fyaw", self.p1.o, self.mech.F_yaw)
        self._register_node_coupling("P2o_Fyaw", self.p2.o, self.mech.F_yaw)
        self._register_node_coupling("P1i_Fpitch", self.p1.i, self.mech.F_pitch)
        self._register_node_coupling("P2i_Fpitch", self.p2.i, self.mech.F_pitch)
        self._register_node_coupling("P1o_Fpitch", self.p1.o, self.mech.F_pitch)
        self._register_node_coupling("P2o_Fpitch", self.p2.o, self.mech.F_pitch)

        # motion to optic coupling: phase coupling on reflection
        self._register_node_coupling("Z_P1o", self.mech.z, self.p1.o)
        self._register_node_coupling("Z_P2o", self.mech.z, self.p2.o)
        self._register_node_coupling("yaw_P1o", self.mech.yaw, self.p1.o)
        self._register_node_coupling("yaw_P2o", self.mech.yaw, self.p2.o)
        self._register_node_coupling("pitch_P1o", self.mech.pitch, self.p1.o)
        self._register_node_coupling("pitch_P2o", self.mech.pitch, self.p2.o)

        # Define typical degrees of freedom for this component
        self.dofs = types.SimpleNamespace()
        self.dofs.z = LocalDegreeOfFreedom(
            f"{self.name}.dofs.z", self.phi, self.mech.z, None
        )
        self.dofs.yaw = LocalDegreeOfFreedom(
            f"{self.name}.dofs.yaw", self.xbeta, self.mech.yaw, 1
        )
        self.dofs.pitch = LocalDegreeOfFreedom(
            f"{self.name}.dofs.pitch", self.ybeta, self.mech.pitch, 1
        )
        self.dofs.F_z = LocalDegreeOfFreedom(
            f"{self.name}.dofs.F_z", self.phi, self.mech.F_z, None, AC_OUT=self.mech.z
        )
        self.dofs.F_yaw = LocalDegreeOfFreedom(
            f"{self.name}.dofs.F_yaw",
            self.xbeta,
            self.mech.F_yaw,
            None,
            AC_OUT=self.mech.yaw,
        )
        self.dofs.F_pitch = LocalDegreeOfFreedom(
            f"{self.name}.dofs.F_pitch",
            self.ybeta,
            self.mech.F_pitch,
            None,
            AC_OUT=self.mech.pitch,
        )

    def optical_equations(self):
        with finesse.symbols.simplification():
            _f_ = Variable("_f_")
            _f0_ = Variable("_f0_")

            r = (self.R.ref) ** 0.5
            t = (self.T.ref) ** 0.5
            phi = self.phi.ref * (1 + _f_ / _f0_) * np.pi / 180

            nr1 = self.ports[0].refractive_index
            nr2 = self.ports[1].refractive_index

            if self._model._settings.phase_config.v2_transmission_phase or nr1 == nr2:
                # old v2 phase on transmission
                # The usual i on transmission and reflections
                # are opposite phase on each side, ignores refractive index
                phi_r1 = 2 * phi
                r1 = r * np.exp(1j * phi_r1)
                r2 = r * np.exp(-1j * phi_r1)
                t12 = t21 = 1j * t
            else:
                # Uses N=-1, Eq.2.25 in Living Reviews in Relativity (2016)
                # 19:3 DOI 10.1007/s41114-016-0002-8
                # beamsplitter transmission phase depends on the reflectivity
                # refractive indices and angle of incidence
                phi_r1 = +2 * phi * nr1
                phi_r2 = -2 * phi * nr2
                phi_t = np.pi / 2 + 0.5 * (phi_r1 + phi_r2)
                r1 = r * np.exp(1j * phi_r1)
                r2 = r * np.exp(1j * phi_r2)
                t12 = t * np.exp(1j * phi_t)
                t21 = t * np.exp(-1j * phi_t)

            if self._model._settings.is_modal:
                return {
                    f"{self.name}.P1i_P1o": r1
                    * (1 - self.misaligned.ref)
                    * Matrix(f"{self.name}.K11"),
                    f"{self.name}.P2i_P2o": r2
                    * (1 - self.misaligned.ref)
                    * Matrix(f"{self.name}.K22"),
                    f"{self.name}.P1i_P2o": t12 * Matrix(f"{self.name}.K12"),
                    f"{self.name}.P2i_P1o": t21 * Matrix(f"{self.name}.K21"),
                }
            else:
                return {
                    f"{self.name}.P1i_P1o": r1 * (1 - self.misaligned.ref),
                    f"{self.name}.P2i_P2o": r2 * (1 - self.misaligned.ref),
                    f"{self.name}.P1i_P2o": t12,
                    f"{self.name}.P2i_P1o": t21,
                }

    def _resymbolise_ABCDs(self):
        # -> reflections
        self.__symbolise_ABCD(self.p1.i, self.p1.o, "x")
        self.__symbolise_ABCD(self.p1.i, self.p1.o, "y")
        self.__symbolise_ABCD(self.p2.i, self.p2.o, "x")
        self.__symbolise_ABCD(self.p2.i, self.p2.o, "y")
        # -> transmissions
        self.__symbolise_ABCD(self.p1.i, self.p2.o, "x")
        self.__symbolise_ABCD(self.p1.i, self.p2.o, "y")
        self.__symbolise_ABCD(self.p2.i, self.p1.o, "x")
        self.__symbolise_ABCD(self.p2.i, self.p1.o, "y")

    @property
    def refractive_index_1(self):
        """Refractive index on size 1 (port 1)"""
        if self.p1.attached_to:
            return self.p1.refractive_index
        else:
            return Constant(1)

    @property
    def refractive_index_2(self):
        """Refractive index on size 2 (port 2)"""

        if self.p2.attached_to:
            return self.p2.refractive_index
        else:
            return Constant(1)

    def __symbolise_ABCD(self, from_node, to_node, direction):

        if self.interaction_type(from_node, to_node) == InteractionType.REFLECTION:
            # reflection
            nr = refractive_index(from_node, symbolic=True)
            if direction == "x":
                Rc = self.Rcx.ref
                ABCD = abcd.mirror_refl_t
            else:
                Rc = self.Rcy.ref
                ABCD = abcd.mirror_refl_s

            # Opposite side of mirror looks inversely curved
            if from_node.port is self.p1:
                M_sym = ABCD(Rc, nr=nr)
            else:
                M_sym = ABCD(-Rc, nr=nr)

        else:  # transmission
            nr1 = self.refractive_index_1
            nr2 = self.refractive_index_2
            if direction == "x":
                Rc = self.Rcx.ref
            else:
                Rc = self.Rcy.ref

            if from_node.port is self.p1:
                M_sym = abcd.mirror_trans(Rc, nr1=nr1, nr2=nr2)
            else:
                M_sym = abcd.mirror_trans(-Rc, nr1=nr2, nr2=nr1)

        key = (from_node, to_node, direction)
        # For mirrors the symbol nr can change when connected up to spaces
        # in a model so here we update the ABCD matrix entry
        if key in self._abcd_matrices:
            self._abcd_matrices[key] = M_sym, np.array(M_sym, dtype=np.float64)
        # Otherwise just register the new ABCD matrix as usual
        else:
            self.register_abcd_matrix(M_sym, (from_node, to_node, direction))

    @property
    def abcd11x(self):
        """Numeric ABCD matrix from port 1 to port 1 in the tangential plane.

        Equivalent to ``mirror.ABCD(1, 1, "x")``.

        :`getter`: Returns a copy of the (numeric) ABCD matrix for this coupling
                   (read-only).
        """
        return self.ABCD(1, 1, "x")

    @property
    def abcd11y(self):
        """Numeric ABCD matrix from port 1 to port 1 in the sagittal plane.

        Equivalent to ``mirror.ABCD(1, 1, "y")``.

        :`getter`: Returns a copy of the (numeric) ABCD matrix for this coupling
                   (read-only).
        """
        return self.ABCD(1, 1, "y")

    @property
    def abcd22x(self):
        """Numeric ABCD matrix from port 2 to port 2 in the tangential plane.

        Equivalent to ``mirror.ABCD(2, 2, "x")``.

        :`getter`: Returns a copy of the (numeric) ABCD matrix for this coupling
                   (read-only).
        """
        return self.ABCD(2, 2, "x")

    @property
    def abcd22y(self):
        """Numeric ABCD matrix from port 2 to port 2 in the sagittal plane.

        Equivalent to ``mirror.ABCD(2, 2, "y")``.

        :`getter`: Returns a copy of the (numeric) ABCD matrix for this coupling
                   (read-only).
        """
        return self.ABCD(2, 2, "y")

    @property
    def abcd12x(self):
        """Numeric ABCD matrix from port 1 to port 2 in the tangential plane.

        Equivalent to ``mirror.ABCD(1, 2, "x")``.

        :`getter`: Returns a copy of the (numeric) ABCD matrix for this coupling
                   (read-only).
        """
        return self.ABCD(1, 2, "x")

    @property
    def abcd12y(self):
        """Numeric ABCD matrix from port 1 to port 2 in the sagittal plane.

        Equivalent to ``mirror.ABCD(1, 2, "y")``.

        :`getter`: Returns a copy of the (numeric) ABCD matrix for this coupling
                   (read-only).
        """
        return self.ABCD(1, 2, "y")

    @property
    def abcd21x(self):
        """Numeric ABCD matrix from port 2 to port 1 in the tangential plane.

        Equivalent to ``mirror.ABCD(2, 1, "x")``.

        :`getter`: Returns a copy of the (numeric) ABCD matrix for this coupling
                   (read-only).
        """
        return self.ABCD(2, 1, "x")

    @property
    def abcd21y(self):
        """Numeric ABCD matrix from port 2 to port 1 in the sagittal plane.

        Equivalent to ``mirror.ABCD(2, 1, "y")``.

        :`getter`: Returns a copy of the (numeric) ABCD matrix for this coupling
                   (read-only).
        """
        return self.ABCD(2, 1, "y")

    def ABCD(
        self,
        from_node,
        to_node,
        direction="x",
        symbolic=False,
        copy=True,
        retboth=False,
        allow_reverse=False,
    ):
        r"""Returns the ABCD matrix of the mirror for the specified coupling.

        In both cases below, the sign of the radius is defined such that :math:`R_c`
        is negative if the centre of the sphere is located in the direction of propagation.

        .. rubric:: Transmission

        .. image:: /images/mirror/abcd_mi.*
            :align: center

        For transmission this is given by,

        .. math::
            M_{t} = \begin{pmatrix}
                        1 & 0 \\
                        \frac{n_2 - n_1}{R_c} & 1
                    \end{pmatrix},

        where :math:`n_2` and :math:`n_1` are the indices of refraction of the spaces
        connected to the mirror and :math:`R_c` is the radius of curvature of the mirror.

        The matrix for transmission in the opposite direction of propagation is identical.

        .. rubric:: Reflection

        .. image:: /images/mirror/abcd_mr.*
            :align: center

        In the case of reflection the matrix is,

        .. math::
            M_{r} = \begin{pmatrix}
                        1 & 0 \\
                        -\frac{2n_1}{R_c} & 1
                    \end{pmatrix}.

        Reflection at the back surface can be described by the same type of matrix by setting
        the :math:`C` element to :math:`C = 2n_2/R_c`.

        See :meth:`.Connector.ABCD` for descriptions of parameters, return values and possible
        exceptions.
        """
        return super().ABCD(
            from_node, to_node, direction, symbolic, copy, retboth, allow_reverse
        )

    def _get_workspace(self, sim):
        """Returns a workspace to use in a simulation."""
        from finesse.simulations.sparse.simulation import SparseMatrixSimulation

        if isinstance(sim, SparseMatrixSimulation):
            from finesse.components.modal.mirror import (
                mirror_carrier_fill,
                mirror_signal_opt_fill,
                mirror_signal_mech_fill,
                mirror_fill_qnoise,
                MirrorWorkspace,
            )

            _, is_changing = self._eval_parameters()

            carrier_refill = (
                sim.is_component_in_mismatch_couplings(self)
                or self in sim.trace_forest
                or sim.carrier.any_frequencies_changing
                or len(is_changing)
            )
            ws = MirrorWorkspace(self, sim)
            ws.imaginary_transmission = self.imaginary_transmission
            # This assumes that nr1/nr2 cannot change during a simulation
            ws.nr1 = refractive_index(self.p1)
            ws.nr2 = refractive_index(self.p2)

            ws.carrier.add_fill_function(mirror_carrier_fill, carrier_refill)

            if sim.signal:
                signal_refill = carrier_refill or (
                    sim.model.fsig.f.is_changing
                    and not (self.phi.value == 0 and self.phi.is_changing)
                )
                ws.signal.add_fill_function(mirror_signal_opt_fill, signal_refill)

                signal_refill = carrier_refill or (sim.model.fsig.f.is_changing)
                ws.signal.add_fill_function(mirror_signal_mech_fill, signal_refill)

            # Initialise the ABCD matrix memory-views
            if sim.is_modal:
                ws.abcd_p1p1_x = self.ABCD(self.p1.i, self.p1.o, "x", copy=False)
                ws.abcd_p1p1_y = self.ABCD(self.p1.i, self.p1.o, "y", copy=False)
                ws.abcd_p2p2_x = self.ABCD(self.p2.i, self.p2.o, "x", copy=False)
                ws.abcd_p2p2_y = self.ABCD(self.p2.i, self.p2.o, "y", copy=False)

                ws.abcd_p1p2_x = self.ABCD(self.p1.i, self.p2.o, "x", copy=False)
                ws.abcd_p1p2_y = self.ABCD(self.p1.i, self.p2.o, "y", copy=False)
                ws.abcd_p2p1_x = self.ABCD(self.p2.i, self.p1.o, "x", copy=False)
                ws.abcd_p2p1_y = self.ABCD(self.p2.i, self.p1.o, "y", copy=False)

                ws.set_knm_info(
                    "P1i_P1o",
                    abcd_x=ws.abcd_p1p1_x,
                    abcd_y=ws.abcd_p1p1_y,
                    nr_from=ws.nr1,
                    nr_to=ws.nr1,
                    is_transmission=False,
                    beta_x=self.xbeta,
                    beta_x_factor=2,
                    beta_y=self.ybeta,
                    beta_y_factor=-2,
                    apply_map=self.surface_map,
                    map_phase_factor=-2 * ws.nr1,
                    map_fliplr=False,
                )
                ws.set_knm_info(
                    "P2i_P2o",
                    abcd_x=ws.abcd_p2p2_x,
                    abcd_y=ws.abcd_p2p2_y,
                    nr_from=ws.nr2,
                    nr_to=ws.nr2,
                    is_transmission=False,
                    beta_x=self.xbeta,
                    beta_x_factor=2,
                    beta_y=self.ybeta,
                    beta_y_factor=2,
                    apply_map=self.surface_map,
                    map_phase_factor=2 * ws.nr2,
                    map_fliplr=True,
                )
                ws.set_knm_info(
                    "P1i_P2o",
                    abcd_x=ws.abcd_p1p2_x,
                    abcd_y=ws.abcd_p1p2_y,
                    nr_from=ws.nr1,
                    nr_to=ws.nr2,
                    is_transmission=True,
                    beta_x=self.xbeta,
                    beta_x_factor=-(1 - ws.nr1 / ws.nr2),
                    beta_y=self.ybeta,
                    beta_y_factor=(1 - ws.nr1 / ws.nr2),
                    apply_map=self.surface_map,
                    map_phase_factor=(ws.nr2 - ws.nr1),
                    map_fliplr=False,
                )
                ws.set_knm_info(
                    "P2i_P1o",
                    abcd_x=ws.abcd_p2p1_x,
                    abcd_y=ws.abcd_p2p1_y,
                    nr_from=ws.nr2,
                    nr_to=ws.nr1,
                    is_transmission=True,
                    beta_x=self.xbeta,
                    beta_x_factor=-(1 - ws.nr2 / ws.nr1),
                    beta_y=self.ybeta,
                    beta_y_factor=-(1 - ws.nr2 / ws.nr1),
                    apply_map=self.surface_map,
                    map_phase_factor=-(ws.nr1 - ws.nr2),
                    map_fliplr=True,
                )

            if sim.signal:
                ws.signal.set_fill_noise_function(NoiseType.QUANTUM, mirror_fill_qnoise)
            return ws



"""Tools for making an arbitrary Finesse object freezable."""


def canFreeze(cls):
    """A function to be used as a decorator for classes which should have the ability to
    be frozen and unfrozen.

    Parameters
    ----------
    cls : object
        Class to make freezable, should never be called directly as
        this function is to be used as a decorator.

    Returns
    -------
    cls : object
        The class with frozen attributes set.

    Examples
    --------
    To make a custom object in Finesse frozen simply use this
    function as a decorator to the class definition::

        from finesse.freeze import canFreeze

        @canFreeze
        class FinesseObject:
            ...
    """

    def _freeze(self):
        self.__dict__["____FROZEN____"] = True

    def _unfreeze(self):
        self.__dict__["____FROZEN____"] = False

    def frozensetattr(self, name, value):
        if (
            "____FROZEN____" in self.__dict__
            and self.__dict__["____FROZEN____"]
            and not hasattr(self, name)
        ):
            # due to python name mangling
            private_name = f"_{self.__class__.__name__}__name"
            if hasattr(self, "name"):
                n = self.name
            elif hasattr(self, private_name):
                n = getattr(self, private_name)
            else:
                n = self.__class__.__name__

            raise TypeError("'%s' does not have attribute called '%s'" % (n, name))

        super(cls, self).__setattr__(name, value)

    cls.__setattr__ = frozensetattr
    cls._freeze = _freeze
    cls._unfreeze = _unfreeze
    return cls


@canFreeze
class Freezable:
    """An object which can be frozen and unfrozen, it is frozen upon construction."""

    def __init__(self):
        """Constructs a new instance of `Freezable` and freezes it."""
        self._freeze()

    def __iter__(self):
        return (val for _, val in self.items())

    def items(self):
        return ((key, val) for key, val in self.__dict__.items() if key[0] != "_")

    def keys(self):
        return [key for key, val in self.items()]

    def values(self):
        return [val for key, val in self.items()]



"""Optical components performing directional suppression of beams."""

import logging
import finesse
import numpy as np

from finesse.components.node import NodeType, NodeDirection
from finesse.components.general import (
    Connector,
    InteractionType,
)
from finesse.parameter import float_parameter
from finesse.utilities import refractive_index
from finesse.symbols import Matrix
from finesse.tracing import abcd

LOGGER = logging.getLogger(__name__)


@float_parameter("S", "Power suppression")
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Isolator(Connector):
    """Represents an isolator optical component with a suppression factor. Suppresses
    the light field transmitted from p2 to p1. The field from p1 to p2 is.

    Parameters
    ----------
    name : str
        Name of newly created isolator.

    S : float
        Power suppression in dB. Defaults to 0
    """

    def __init__(self, name, S=0.0):
        super().__init__(name)
        self.S = S

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p2", NodeType.OPTICAL)
        self.p2._add_node("i", NodeDirection.INPUT)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        # optic to optic couplings
        self._register_node_coupling(
            "P1i_P2o",
            self.p1.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
        )

    def _resymbolise_ABCDs(self):
        M_sym = abcd.none()
        for direction in ["x", "y"]:
            # Matrices same for both node couplings
            self.register_abcd_matrix(
                M_sym,
                (self.p1.i, self.p2.o, direction),
                (self.p2.i, self.p1.o, direction),
            )

    @property
    def suppression_factor(self) -> float:
        """Factor by which the light field transmitted from p2 to p1 is supressed."""
        if self._model and self._model.is_built:
            S = self._eval_parameters()[0]["S"]
        else:
            S = self.S.ref
        return 10 ** (-S / 20)

    def optical_equations(self):
        with finesse.symbols.simplification():
            S = self.suppression_factor
            if self._model._settings.is_modal:
                return {
                    f"{self.name}.P1i_P2o": S * Matrix(f"{self.name}.K12"),
                    f"{self.name}.P2i_P1o": S * Matrix(f"{self.name}.K21"),
                }
            else:
                return {
                    f"{self.name}.P1i_P2o": S,
                    f"{self.name}.P2i_P1o": S,
                }

    def _get_workspace(self, sim):
        from finesse.components.modal.isolator import IsolatorWorkspace

        _, is_changing = self._eval_parameters()
        refill = sim.is_component_in_mismatch_couplings(self) or len(is_changing)

        ws = IsolatorWorkspace(self, sim)
        # This assumes that nr1/nr2 cannot change during a simulation
        ws.nr1 = refractive_index(self.p1)
        ws.nr2 = refractive_index(self.p2)
        ws.carrier.add_fill_function(self._fill_carrier, refill)
        if sim.signal:
            ws.signal.add_fill_function(self._fill_signal, refill)

        if sim.is_modal:
            ws.set_knm_info(
                "P1i_P2o", nr_from=ws.nr1, nr_to=ws.nr2, is_transmission=True
            )
            ws.set_knm_info(
                "P2i_P1o", nr_from=ws.nr2, nr_to=ws.nr1, is_transmission=True
            )

        return ws

    def _fill_carrier(self, ws):
        self._fill_matrix(ws, ws.sim.carrier, ws.carrier.connections)

    def _fill_signal(self, ws):
        self._fill_matrix(ws, ws.sim.signal, ws.signal.connections)

    def _fill_matrix(self, ws, mtx, connections):
        for freq in mtx.optical_frequencies.frequencies:
            with mtx.component_edge_fill3(
                ws.owner_id, connections.P1i_P2o_idx, freq.index, freq.index
            ) as mat:
                mat[:] = ws.K12.data

            with mtx.component_edge_fill3(
                ws.owner_id,
                connections.P2i_P1o_idx,
                freq.index,
                freq.index,
            ) as mat:
                np.multiply(self.suppression_factor, ws.K21.data, out=mat[:])



"""Extract extra dependencies from `pyproject.toml`.

Call with the path to `pyproject.toml` and the extras section name(s) to extract as
arguments.

This is used by the CI to install build dependencies without the need for a separate
`requirements-build.txt` file.

Note that we are using a regular expression here, since a toml parsers was only added to
the python standard library in 3.11, for older version you would have to install a
separate package to run this script, complicating the build process.
"""

from __future__ import annotations

import argparse
import re
from pathlib import Path
from typing import Generator


def yield_reqs(reqs: str) -> Generator[str, None, None]:
    for line in reqs.splitlines():
        line = line.strip()
        if not line.startswith('"'):
            continue
        line = line[line.index('"') + 1 :]
        line = line[: line.index('"')]
        yield line


def extract_reqs(pyproject_path: Path, section: str) -> str:
    with open(pyproject_path, "r") as f:
        cfg = f.read()
    match = re.search(
        r"""
        (\[project\.optional-dependencies\]  # Optional dependencies section
        .*?)  # any characters inside of it
        (\Z|^\[)  # until we reach the end of the string or a new section
        """,
        cfg,
        flags=re.S | re.M | re.X,
    )
    if match is None:
        raise ValueError("Project optional dependencies not found in pyproject.toml")
    optional_deps = match.group()

    section_matches = re.finditer(
        pattern=r"""(^\w.*?)  # name of the section
                    \s?=\s?   # equal sign possibly surrounded by spaces
                    (\[.*?\]) # any multiline string in between square brackets
                """,
        string=optional_deps,
        flags=re.S | re.M | re.X,
    )
    for section_match in section_matches:
        name, reqs = section_match.groups()
        if name == section:
            return reqs
    raise ValueError(f"Section {section} not found")


def main(pyproject_path: Path, section: str) -> None:
    for line in yield_reqs(extract_reqs(pyproject_path, section)):
        print(line)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Extra extra requirments section from pyproject.toml file"
    )
    parser.add_argument("pyproject_path", type=Path, help="Path to pyproject.toml file")
    parser.add_argument(
        "section", type=str, help="Section of optional dependencies to extract"
    )
    args = parser.parse_args()
    main(pyproject_path=args.pyproject_path, section=args.section)



import pytest
from testutils.tokens import IMPLICITLINEEND, ENDMARKER, NAME


@pytest.mark.parametrize(
    "string",
    (
        "m1",
        "m2.T",
        "m3.phi",
        "m4.p1.o.q",
    ),
)
def test_reference(tokenizer, string):
    assert list(tokenizer.tokenize(string)) == [
        NAME(1, 1, string),
        IMPLICITLINEEND(1, 1 + len(string)),
        ENDMARKER(2),
    ]



"""Test parameter errors."""

import pytest
from finesse.components import Laser, Mirror, Space
from finesse.detectors import PowerDetector
from finesse.exceptions import (
    ContextualTypeError,
    ContextualValueError,
)
from testutils.text import escape_full


def test_type_error(model):
    """Incorrect parameter types should raise an error that offers allowed types."""
    laser = Laser("l1")

    with pytest.raises(
        ContextualTypeError,
        match=escape_full("\npd node: invalid type (expected 'Node', got 'Laser')"),
    ):
        model.add(PowerDetector("pd", laser))


def test_value_error(model):
    laser = Laser("l1")
    mirror = Mirror("m1")

    with pytest.raises(
        ContextualValueError,
        match=(
            r"\n'portB': invalid value \<Port m1.mech Type=NodeType.MECHANICAL @ "
            r"(.*)\> \(must be an optical port for s1.portB\)"
        ),
    ):
        model.add(laser)
        model.add(mirror)
        model.add(Space("s1", portA=laser.p1, portB=mirror.mech))


# changed the behaviour when renaming select_modes to modes
@pytest.mark.xfail
def test_multi_value_error(model):
    with pytest.raises(
        ContextualValueError,
        match=escape_full(
            "\n'select' and 'maxtem': invalid values (cannot both be empty)"
        ),
    ):
        model.modes()



"""KatScript specification tests."""


def test_register_element(spec, fake_element_adapter_factory, fake_element_cls):
    adapter = fake_element_adapter_factory(fake_element_cls)

    assert adapter.full_name not in spec.elements
    spec.register_element(adapter)
    assert adapter.full_name in spec.elements



from finesse.frequency cimport frequency_info_t
from finesse.cymath.cmatrix cimport SubCCSView, SubCCSView1DArray, SubCCSView2DArray
from finesse.knm cimport KnmMatrix
from finesse.cymath cimport complex_t
from finesse.simulations.base cimport ModelSettings, NodeBeamParam
from finesse.element_workspace cimport BaseCValues
from finesse.components.workspace cimport ConnectorWorkspace, FillFuncWrapper
from finesse.components.modal.workspace cimport KnmConnectorWorkspace

from finesse.cyexpr cimport (
    cy_expr,
    cy_expr_new,
    cy_expr_init,
    cy_expr_free,
    cy_expr_eval,
)

from cpython.ref cimport PyObject

import numpy as np
cimport numpy as np

cdef class MirrorValues(BaseCValues):
    cdef public:
        double R
        double T
        double L
        double phi
        double Rcx
        double Rcy
        double xbeta
        double ybeta
        double misaligned

cdef struct PyObject1DArray:
    PyObject** ptr
    Py_ssize_t size

cdef struct PyObject2DArray:
    PyObject*** ptr
    Py_ssize_t size1
    Py_ssize_t size2

cdef struct mirror_optical_connections:
    PyObject** P1i_P1o
    PyObject** P1i_P2o
    PyObject** P2i_P1o
    PyObject** P2i_P2o

cdef struct mirror_signal_connections:
    PyObject*** P1i_Fz
    PyObject*** P1o_Fz
    PyObject*** P2i_Fz
    PyObject*** P2o_Fz
    PyObject*** P1i_Fyaw
    PyObject*** P1o_Fyaw
    PyObject*** P2i_Fyaw
    PyObject*** P2o_Fyaw
    PyObject*** P1i_Fpitch
    PyObject*** P1o_Fpitch
    PyObject*** P2i_Fpitch
    PyObject*** P2o_Fpitch
    PyObject*** Z_P1o
    PyObject*** Z_P2o
    PyObject*** yaw_P1o
    PyObject*** yaw_P2o
    PyObject*** pitch_P1o
    PyObject*** pitch_P2o

cdef class MirrorOpticalConnections:
    cdef public:
        int P1i_P1o_idx
        int P2i_P2o_idx
        int P1i_P2o_idx
        int P2i_P1o_idx

    cdef readonly:
        SubCCSView1DArray P1i_P1o
        SubCCSView1DArray P2i_P2o
        SubCCSView1DArray P1i_P2o
        SubCCSView1DArray P2i_P1o
    cdef:
        mirror_optical_connections opt_conn_ptrs

cdef class MirrorSignalConnections(MirrorOpticalConnections):
    cdef public:
        int P1i_Fz_idx, P2i_Fz_idx, P1o_Fz_idx, P2o_Fz_idx
        int P1i_Fyaw_idx, P2i_Fyaw_idx, P1o_Fyaw_idx, P2o_Fyaw_idx
        int P1i_Fpitch_idx, P2i_Fpitch_idx, P1o_Fpitch_idx, P2o_Fpitch_idx
        int Z_P1o_idx, Z_P2o_idx, yaw_P1o_idx, yaw_P2o_idx, pitch_P1o_idx, pitch_P2o_idx

    cdef readonly:
        SubCCSView2DArray P1i_Fz
        SubCCSView2DArray P1o_Fz
        SubCCSView2DArray P2i_Fz
        SubCCSView2DArray P2o_Fz
        SubCCSView2DArray P1i_Fyaw
        SubCCSView2DArray P1o_Fyaw
        SubCCSView2DArray P2i_Fyaw
        SubCCSView2DArray P2o_Fyaw
        SubCCSView2DArray P1i_Fpitch
        SubCCSView2DArray P1o_Fpitch
        SubCCSView2DArray P2i_Fpitch
        SubCCSView2DArray P2o_Fpitch
        SubCCSView2DArray Z_P1o
        SubCCSView2DArray Z_P2o
        SubCCSView2DArray yaw_P1o
        SubCCSView2DArray yaw_P2o
        SubCCSView2DArray pitch_P1o
        SubCCSView2DArray pitch_P2o
    cdef:
        mirror_signal_connections sig_conn_ptrs

cdef class MirrorWorkspace(KnmConnectorWorkspace):
    cdef public:
        complex_t field_to_F
        complex_t z_to_field
        bint imaginary_transmission

        # Complete scattering matrices for each propagation direction
        KnmMatrix K11
        KnmMatrix K12
        KnmMatrix K21
        KnmMatrix K22

        KnmMatrix K_yaw_sig, K_pitch_sig

        # NOTE (sjr) The above should remain as the full scattering matrices
        #            (i.e. composite over each scattering type) whilst extra
        #            KnmMatrix objects should be stored here for the different
        #            types (i.e. K11_map, K11_aperture, etc.)

        # Arrays of scattering losses for each mode coupling
        double[::1] K11_loss
        double[::1] K12_loss
        double[::1] K21_loss
        double[::1] K22_loss

        # Node indices for the carrier simulation
        Py_ssize_t car_p1i_idx, car_p1o_idx
        Py_ssize_t car_p2i_idx, car_p2o_idx

        Py_ssize_t car_p1o_rhs_idx, car_p1i_rhs_idx
        Py_ssize_t car_p2o_rhs_idx, car_p2i_rhs_idx
        Py_ssize_t car_p_num_hom

        # Refractive indices of adjacent spaces
        double nr1
        double nr2

        # ABCD matrix views
        # -> reflection
        double[:, ::1] abcd_p1p1_x
        double[:, ::1] abcd_p1p1_y
        double[:, ::1] abcd_p2p2_x
        double[:, ::1] abcd_p2p2_y
        # -> transmission
        double[:, ::1] abcd_p1p2_x
        double[:, ::1] abcd_p1p2_y
        double[:, ::1] abcd_p2p1_x
        double[:, ::1] abcd_p2p1_y
    cdef:
        MirrorValues mv
        MirrorOpticalConnections mcc
        MirrorValues cvalues
        MirrorSignalConnections mcs
        bint z_signal_enabled, yaw_signal_enabled, pitch_signal_enabled

        # Indices (in sim._c_node_info and sim.trace) of nodes
        Py_ssize_t P1i_id
        Py_ssize_t P1o_id
        Py_ssize_t P2i_id
        Py_ssize_t P2o_id

        # Only elements C of ABCD matrices can change, store pointers
        # to compiled changing expression for these elements here
        # -> size 8 as there are 8 different coupling, plane combos
        # -> order of these corresponds exactly to order of matrix views above
        cy_expr* sym_abcd_Cs[8]

        # Direct pointer access to C element of each ABCD matrix for
        # convenience in updating
        double* abcd_Cs[8]

        frequency_info_t *z_mech_freqs
        Py_ssize_t z_mech_freqs_size
        frequency_info_t *yaw_mech_freqs
        Py_ssize_t yaw_mech_freqs_size
        frequency_info_t *pitch_mech_freqs
        Py_ssize_t pitch_mech_freqs_size

    cpdef update_parameter_values(self)



"""Space tests."""

import pytest
import finesse
import numpy as np


@pytest.fixture
def simple_cavity_model():
    def getmodel(user_gouy=""):
        model = finesse.Model()
        model.parse(
            f"""
            l l1
            s s0 l1.p1 m1.p1
            m m1 Rc=-0.5
            s cav m1.p2 m2.p1 L=0.25 {user_gouy}
            m m2 Rc=0.5
            cav cavity m1.p2.o
            """
        )

        return model

    return getmodel


def test_getter_returns_default_gouy_without_user(simple_cavity_model):
    model = simple_cavity_model()

    assert model.spaces.s0.gouy_x == 0.0
    assert model.spaces.s0.gouy_y == 0.0
    assert model.spaces.cav.gouy_x == 0.0
    assert model.spaces.cav.gouy_y == 0.0


def test_getter_returns_user_gouy_with_user(simple_cavity_model):
    model = simple_cavity_model("user_gouy_x=1 user_gouy_y=2")

    assert model.spaces.s0.gouy_x == 0.0
    assert model.spaces.s0.gouy_y == 0.0
    assert model.spaces.cav.gouy_x == 1.0
    assert model.spaces.cav.gouy_y == 2.0


def test_user_gouy_is_none_without_user(simple_cavity_model):
    model = simple_cavity_model()

    assert model.spaces.s0.user_gouy_x.value is None
    assert model.spaces.s0.user_gouy_y.value is None
    assert model.spaces.cav.user_gouy_x.value is None
    assert model.spaces.cav.user_gouy_y.value is None


def test_user_gouy_is_none_with_user(simple_cavity_model):
    model = simple_cavity_model("user_gouy_x=1 user_gouy_y=2")

    assert model.spaces.s0.user_gouy_x.value is None
    assert model.spaces.s0.user_gouy_y.value is None
    assert model.spaces.cav.user_gouy_x.value == 1.0
    assert model.spaces.cav.user_gouy_y.value == 2.0


def test_gouy_is_being_set_during_run(simple_cavity_model):
    model = simple_cavity_model()
    model.run()

    assert np.isclose(model.spaces.cav.gouy_x, 60.0)
    assert np.isclose(model.spaces.cav.gouy_y, 60.0)


def test_user_gouy_is_being_set_during_run(simple_cavity_model):
    model = simple_cavity_model("user_gouy_x=1 user_gouy_y=2")
    model.run()

    assert np.isclose(model.spaces.cav.gouy_x, 1.0)
    assert np.isclose(model.spaces.cav.gouy_y, 2.0)



import pytest
from finesse.script import parse, unparse


@pytest.fixture
def unparsed_script():
    def getmodel(user_gouy=""):
        return unparse(
            parse(
                f"""
                l l1
                s s0 l1.p1 m1.p1 {user_gouy}
                m m1
                """
            )
        )

    return getmodel


def test_no_user_gouy(unparsed_script):
    assert "user_gouy" not in unparsed_script()
    assert "user_gouy" not in unparsed_script("user_gouy_x=None user_gouy_y=None")


def test_with_user_gouy(unparsed_script):
    assert "user_gouy_x=1" in unparsed_script("user_gouy_x=1")
    assert "user_gouy_x=1" in unparsed_script("user_gouy_x=1 user_gouy_y=1")

    assert "user_gouy_y=1" in unparsed_script("user_gouy_y=1")
    assert "user_gouy_y=1" in unparsed_script("user_gouy_x=1 user_gouy_y=1")



#cython: boundscheck=False, wraparound=False, initializedcheck=False

"""Scattering matrix data structure and associated functions
for constructing different formats of this object."""

cimport numpy as np
import numpy as np

from finesse.cymath.math cimport sqrt
from finesse.cymath.complex cimport cnorm
from finesse.cymath.homs cimport field_index
from finesse.cymath.complex cimport carg # Standard complex.h functions
from finesse.cymath.complex cimport crotate, crotate2, cnorm
from finesse.cymath.math cimport sqrt, cos, sin # Standard math.h functions
from finesse.cymath.math cimport msign

import copy
import cython


cdef class KnmMatrix:
    """
    Higher-Order-Mode scattering matrix container. Essentially a wrapper
    around a 2D NumPy array with methods for conveniently accessing specific
    couplings and plotting the matrix as a color-mesh.

    The underlying :class:`numpy.ndarray` object can be accessed via the ``data``
    attribute.

    Construction of KnmMatrix objects should, generally, not be performed manually. Objects
    of this type are the return type of the general scattering matrix computing routines,
    see :func:`.make_scatter_matrix`. To make a KnmMatrix object from a pre-existing
    matrix of complex coupling coefficients, use :meth:`.KnmMatrix.from_buffer`.

    Parameters
    ----------
    modes : array-like
        A 2D array, or memory-view to the array, of the mode indices
        associated with the scattering matrix.

    comp_name : str, optional; default: ""
        Name of the component associated with the matrix (if any).

    kdir : str, optional; default: ""
        A string representing the coupling direction of the matrix, e.g.
        "11" for a reflection at the first surface of a mirror-like optic.
    """

    def __init__(self, const int[:, ::1] modes, comp_name="", kdir=""):
        if modes.shape[0] == 0 and modes.shape[1] == 0:
            raise Exception(f"Empty mode list provided: {np.array(modes)}")

        if comp_name:
            self.name = f"{comp_name}.K{kdir}"
        else:
            self.name = f"K{kdir}"

        self.modes_view = modes
        self.homs = self.modes = np.asarray(self.modes_view)
        matrix = np.eye(self.modes_view.shape[0], dtype=np.complex128)
        self.data = matrix
        self.data_view = self.data
        self.mtx.ptr = <complex_t*>&self.data_view[0,0]
        self.mtx.size1 = self.data_view.shape[0]
        self.mtx.size2 = self.data_view.shape[1]
        self.mtx.stride1 = self.data_view.strides[0]//sizeof(complex_t)
        self.mtx.stride2 = self.data_view.strides[1]//sizeof(complex_t)

    @staticmethod
    def from_buffer(
        np.ndarray[complex_t, ndim=2, mode="c"] buffer,
        const int[:, ::1] modes,
        **kwargs,
    ):
        """Construct a KnmMatrix object from a pre-existing 2D array buffer.

        This method is useful for creating KnmMatrix objects from scattering matrix
        data which have been computed elsewhere. One of the most common use-cases
        is when using :class:`.KnmDetector` objects in "matrix-mode"; see the documentation
        of this class for an example.

        Parameters
        ----------
        buffer : :class:`numpy.ndarray`
            A 2D array of complex values corresponding to the underlying
            buffer of scattering matrix data.

        modes : array-like
            A 2D array, or memory-view to the array, of the mode indices
            associated with the scattering matrix.

        kwargs : keyword arguments, optional
            Additional args to pass to constructor of KnmMatrix.

        Returns
        -------
        kmat : :class:`.KnmMatrix`
            The KnmMatrix wrapper around the array buffer.
        """
        cdef Py_ssize_t N = modes.shape[0]
        if buffer[:].shape[0] != N or buffer[:].shape[1] != N:
            bshape = (buffer[:].shape[0], buffer[:].shape[1])
            raise ValueError(
                f"Error in KnmMatrix.from_buffer:\n"
                f"    Shape of matrix buffer {bshape} inconsistent "
                f"with number of modes ({N})"
            )

        cdef KnmMatrix kmat = KnmMatrix(modes, **kwargs)
        kmat.data = buffer
        kmat.data_view = buffer
        kmat.mtx.ptr = <complex_t*>&kmat.data_view[0,0]
        kmat.mtx.size1 = kmat.data_view.shape[0]
        kmat.mtx.size2 = kmat.data_view.shape[1]
        kmat.mtx.stride1 = kmat.data_view.strides[0]//sizeof(complex_t)
        kmat.mtx.stride2 = kmat.data_view.strides[1]//sizeof(complex_t)
        return kmat

    @staticmethod
    def of_zeros(const int[:, ::1] modes, **kwargs):
        """Constuct a KnmMatrix where each element is (complex) zero.

        Parameters
        ----------
        modes : array-like
            A 2D array, or memory-view to the array, of the mode indices
            associated with the scattering matrix.

        kwargs : keyword arguments, optional
            Additional args to pass to constructor of KnmMatrix.

        Returns
        -------
        kmat : :class:`.KnmMatrix`
            The KnmMatrix object consisting of a matrix of zeroes.
        """
        cdef KnmMatrix kmat = KnmMatrix(modes, **kwargs)
        kmat.data[:] = 0 + 0j
        return kmat

    def __str__(self):
        cdef:
            Py_ssize_t i, j
            int n1, m1, n2, m2
            Py_ssize_t N = self.modes_view.shape[0]

        s = ""
        for i in range(N):
            n1 = self.modes_view[i][0]
            m1 = self.modes_view[i][1]
            for j in range(N):
                n2 = self.modes_view[j][0]
                m2 = self.modes_view[j][1]

                s += (f"{self.name}[{j}][{i}]: {n1} {m1} "
                      f"-> {n2} {m2} = {self.data_view[j][i]}\n")

        return s

    cdef (Py_ssize_t, Py_ssize_t) field_indices_from(self, key):
        cdef:
            Py_ssize_t i = 0
            Py_ssize_t j = 0
            int n1, m1, n2, m2
            Py_ssize_t N = self.modes_view.shape[0]

            Py_ssize_t key_size
            str skey
            str mode1, mode2
            list smodes

        if isinstance(key, str):
            skey = "".join(key.split()) # strip all whitespace from the key
            key_size = len(skey)
            smodes = skey.split("->")
            if not smodes or len(smodes) > 2:
                i = j = N
            else:
                # Key is a str of format "nmn'm'"
                if len(smodes) == 1:
                    if key_size != 4:
                        i = j = N
                    else:
                        mode1, mode2 = smodes[0][:2], smodes[0][2:]

                # Key is a str of format "nm->n'm'"
                elif len(smodes) == 2:
                    if key_size != 6:
                        i = j = N
                    else:
                        mode1, mode2 = smodes

                if not i and not j:
                    # Have to convert to python int class
                    # first before setting C ints
                    nn1 = int(mode1[0])
                    mm1 = int(mode1[1])
                    nn2 = int(mode2[0])
                    mm2 = int(mode2[1])
                    n1 = nn1
                    m1 = mm1
                    n2 = nn2
                    m2 = mm2

                    i = field_index(n1, m1, self.modes_view)
                    j = field_index(n2, m2, self.modes_view)

        else:
            key_size = len(key)
            if key_size == 2: # getting [i, j] element directly
                i, j = key
            elif key_size == 4: # getting nm -> n'm' coupling
                n1, m1, n2, m2 = key

                i = field_index(n1, m1, self.modes_view)
                j = field_index(n2, m2, self.modes_view)
            else:
                i = j = N

        return i, j

    def __getitem__(self, key):
        cdef:
            Py_ssize_t i, j
            Py_ssize_t N = self.modes_view.shape[0]

        i, j = self.field_indices_from(key)

        if i >= N or j >= N:
            raise ValueError(f"Invalid or non-existent key {key}")

        return self.data_view[j][i]

    def get(self, key, default=None):
        """Retrieves the coupling coefficient corresponding to the `key`, returning
        `default` if this `key` is invalid.

        Parameters
        ----------
        key : str or integer sequence
            A string of the format:

                - "nm->n'm'" or,
                - "nmn'm'",

            where n, m, n' and m' are all convertible to integers

            Or an integer sequence of:

                - length 2 -- i.e. an i, j pair corresponding to the field indices.
                - or length 4 -- i.e. a n1, m1, n2, m2 coupling.

        default : Any, optional
            The default value returned if the `key` is invalid.

        Returns
        -------
        out : complex
            The value of the coupling coefficient associated with `key`.
        """
        cdef:
            Py_ssize_t i, j
            Py_ssize_t N = self.modes_view.shape[0]

        i, j = self.field_indices_from(key)

        if i >= N or j >= N:
            return default

        return self.data_view[j][i]

    cdef complex_t coupling(self, int n1, int m1, int n2, int m2) noexcept nogil:
        cdef:
            Py_ssize_t i = field_index(n1, m1, self.modes_view)
            Py_ssize_t j = field_index(n2, m2, self.modes_view)

        return self.data_view[j][i]

    def plot(
            self, mode="amplitude", log=False, deg=False, cmap=None, show=True, filename=None,
    ):
        """Plots the coupling coefficient matrix as a color-mesh.

        Parameters
        ----------
        mode : str, optional
            Specifier for which attribute of the coupling coefficient to
            plot. This can be one of 'amplitude', 'phase', 'real', 'imag',
            'amplitude/phase', or 'real/imag'. Defaults to 'amplitude'.

        log : bool, optional
            Whether the log of the data should be plotted. This is only used for
            'amplitude' and is ignored for all other options. Defaults to False.

        deg : bool, optional
            Whether the data should be plotted in degrees or radians. This is only
            used for 'phase' and is ignored for all other options. Defaults to False.

        cmap : str, matplotlib colormap, optional
            Colormap to use. Defaults to the default colormap loaded in
            matplotlib.rcParams.

        show : bool, optional
            Whether to show plot or not. Defaults to true.

        filename : str, path, optional
            The name of a file to save the figure to. Defaults to None so
            that no file is saved.

        Returns
        -------
        fig : matplotlib figure
            A handle to the figure.
        """
        import matplotlib.pyplot as plt
        import matplotlib.colors as colors
        from finesse.plotting.tools import add_colorbar
        from functools import partial

        knm = self.data[:]

        valid_modes = [
            "amplitude", "phase", "real", "imag", "amplitude/phase", "real/imag",
        ]
        if mode not in valid_modes:
            raise ValueError(
                f"Invalid mode={mode} argument. Must be one of {valid_modes}"
            )
        modes = mode.split("/")
        nplots = len(modes)

        func_map = {
            "amplitude": np.abs,
            "phase": partial(np.angle, deg=deg),
            "real": np.real,
            "imag": np.imag,
        }
        label_map = {
            "amplitude": "abs",
            "phase": f"phase [{'deg' if deg else 'rad'}]",
            "real": "real",
            "imag": "imag",
        }

        fig = plt.figure()
        gs = fig.add_gridspec(nplots, 1, hspace=0.04 * (nplots - 1))
        ax_top = fig.add_subplot(gs[0])
        if nplots == 2:
            ax_bot = fig.add_subplot(gs[1], sharex=ax_top)

        numrows, numcols = knm.shape
        hom_ticks = []
        for i in range(self.modes_view.shape[0]):
            hom_ticks.append(f"{self.modes_view[i][0]}{self.modes_view[i][1]}")
        A = np.arange(1, self.modes_view.shape[0] + 1) - 0.5

        def make_subplot(ax, mode):
            func = func_map[mode]
            label = label_map[mode]

            if log and mode == "amplitude":
                norm = colors.LogNorm()
                # Make sure any zero couplings get displayed as black rather than white
                cmap_handle = copy.copy(plt.colormaps.get_cmap(cmap))
                cmap_handle.set_bad((0,0,0))
            else:
                norm = None
                cmap_handle = cmap

            cax = ax.pcolormesh(func(knm), cmap=cmap_handle, norm=norm)
            add_colorbar(cax, label=label)

            ax.set_xticks(A)
            ax.set_yticks(A)
            ax.set_xticklabels(hom_ticks)
            ax.set_yticklabels(hom_ticks)

            ax.set_xlim(None, np.max(A) + 0.5)
            ax.set_ylim(None, np.max(A) + 0.5)

            def format_coord(x, y):
                col = int(np.floor(x))
                row = int(np.floor(y))

                if col >= 0 and col < numcols and row >= 0 and row < numrows:
                    z = knm[row, col]
                    return (
                        f"nm={hom_ticks[col]}, n'm'={hom_ticks[row]}, z={z:.4g}"
                        f" -> {label}(z) = {func(z):.4g}"
                    )

                return None

            ax.format_coord = format_coord
            # ax.set_aspect("equal")

        make_subplot(ax_top, mode=modes[0])
        if nplots == 2:
            make_subplot(ax_bot, mode=modes[1])
            # don't label the top xaxis since the bottom one will have them
            plt.setp(ax_top.get_xticklabels(), visible=False)

        figsize = fig.get_size_inches()
        fig.set_size_inches(figsize[0], figsize[1] * nplots)
        fig.tight_layout()

        if filename is not None:
            # need bbox_inches here even after tight_layout has been called
            fig.savefig(filename, bbox_inches="tight")

        if show:
            plt.show()

        return fig


cpdef make_unscaled_X_scatter_knm_matrix(int[:, ::1] modes) :
    """ This method returns an unscaled KnmMatrix object that represents
    a distortion from the integral:

    iint_inf U_nm(x,y) x U_n'm'(x,y) dx dy

    This essentially scatters modes by mode index n ± 1. There are some scalings
    proportional to w(x)*exp(±1j*Gouy(z)) missing as these can be applied as single
    scalars in addition to this matrix if needed.

    Returns
    -------
    KnmMatrix
    """
    cdef int i, j, n_, m_, n, m
    cdef KnmMatrix k_yaw = KnmMatrix.of_zeros(modes)
    i = j = -1
    for n_, m_ in modes: # from
        i += 1
        j = -1
        for n, m in modes: # to
            j += 1
            if m == m_:
                if n + 1 == n_:
                    # matrix is hermitian
                    k_yaw.data_view[j, i] = sqrt(n+1)/2
                    k_yaw.data_view[i, j] = sqrt(n+1)/2

    return k_yaw


cpdef make_unscaled_Y_scatter_knm_matrix(int[:, ::1] modes) :
    """ This method returns an unscaled KnmMatrix object that represents
    a distortion from the integral:

    iint_inf U_nm(x,y) y U_n'm'(x,y) dx dy

    This essentially scatters modes by mode index m ± 1. There are some scalings
    proportional to w(x)*exp(±1j*Gouy(z)) missing as these can be applied as single
    scalars in addition to this matrix if needed.

    Returns
    -------
    KnmMatrix
    """
    cdef int i, j, n_, m_, n, m
    cdef KnmMatrix k_pitch = KnmMatrix.of_zeros(modes)
    i = j = -1
    for n_, m_ in modes: # from
        i += 1
        j = -1
        for n, m in modes: # to
            j += 1
            if n == n_:
                if m + 1 == m_:
                    # matrix is hermitian
                    k_pitch.data_view[j, i] = sqrt(m+1)/2
                    k_pitch.data_view[i, j] = sqrt(m+1)/2

    return k_pitch


cdef void knm_loss(const complex_t* knm_mat, double* out, Py_ssize_t N) noexcept nogil:
    """Compute per mode losses due to scattering.

    Each entry in out is then 1 - \sum_{n'm'} |k_{n'm'nm}|^2 for the mode nm.
    """
    cdef:
        Py_ssize_t i, j

    for i in range(N):
        out[i] = 1.0
        for j in range(N):
            out[i] -= cnorm(knm_mat[j*N + i])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void c_zero_tem00_phase(
    const complex_t[:, ::1] knm_mat,
    complex_t[:, ::1] out
) noexcept nogil:
    """Rotates all coupling coefficients in the matrix `knm_mat`
    by the phase of the zeroth (00 -> 00) coupling coefficient.

    This has the effect of zeroing the phase of the 00 -> 00 coupling
    coefficient, resulting in cavities being resonant for the 00 mode
    by default if this function (along with zeroing of space Gouy phase
    for 00 mode) is applied."""
    cdef:
        Py_ssize_t i, j
        Py_ssize_t N = knm_mat.shape[0]
        double phase_k0000 = carg(knm_mat[0][0])
        double cph = cos(phase_k0000)
        double sph = -sin(phase_k0000)

    for i in range(N):
        for j in range(N):
            out[i][j] = crotate2(knm_mat[i][j], cph, sph)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void c_flip_odd_horizontal(
    DenseZMatrix *mat,
    const int[:, ::1] homs
) noexcept nogil:
    """Flips the sign of all odd couplings in the sagittal plane."""
    cdef:
        Py_ssize_t i, j
        int n2
        Py_ssize_t N = mat.size1

    for i in range(N):
        for j in range(N):
            n2 = homs[j][0]
            if msign(n2) == -1:
                mat.ptr[j*mat.stride1 + i*mat.stride2] *= -1


def reverse_gouy_phases(
    x_gouy1, y_gouy1,
    x_gouy2, y_gouy2,
    knm_matrix
):
    """
    Adjust the phase of all coupling coefficients in the matrix `knm_mat` with
    respect to the Gouy phases.

    This is required for :math:`k_{nmn'm'}` calculations because in Finesse the
    Gouy phase is added explicitly to the amplitude coefficients in a
    :class:`.Space` whereas the coupling coefficients are derived using a
    formula in which the Gouy phase resides in the equation for the spatial
    profile.
    """
    c_reverse_gouy_phases(
        x_gouy1, y_gouy1,
        x_gouy2, y_gouy2,
        knm_matrix.data,
        knm_matrix.modes,
        knm_matrix.data
    )


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void c_reverse_gouy_phases(
    double x_gouy1, double y_gouy1,
    double x_gouy2, double y_gouy2,
    const complex_t[:, ::1] knm_mat,
    const int[:, ::1] homs,
    complex_t[:, ::1] out
) noexcept nogil:
    """
    Adjust the phase of all coupling coefficients in the matrix `knm_mat` with
    respect to the Gouy phases.

    This is required for :math:`k_{nmn'm'}` calculations because in Finesse the
    Gouy phase is added explicitly to the amplitude coefficients in a
    :class:`.Space` whereas the coupling coefficients are derived using a
    formula in which the Gouy phase resides in the equation for the spatial
    profile.
    """
    cdef:
        Py_ssize_t i, j
        int n1, m1, n2, m2
        Py_ssize_t N = knm_mat.shape[0]

    for i in range(N):
        n1 = homs[i][0]
        m1 = homs[i][1]
        for j in range(N):
            n2 = homs[j][0]
            m2 = homs[j][1]

            out[j][i] = rev_gouy(
                x_gouy1, y_gouy1,
                x_gouy2, y_gouy2,
                knm_mat[j][i],
                n1, m1, n2, m2,
            )


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef complex_t rev_gouy(
    double x_gouy1,
    double y_gouy1,
    double x_gouy2,
    double y_gouy2,
    complex_t k,
    int n1, int m1, int n2, int m2,
) noexcept nogil:
    cdef:
        double phase1, phase2

    phase1 = (n1 + 0.5) * x_gouy1 + (m1 + 0.5) * y_gouy1
    phase2 = (n2 + 0.5) * x_gouy2 + (m2 + 0.5) * y_gouy2

    return crotate(k, phase2 - phase1)



import pytest
from testutils.tokens import COMMENT, IMPLICITLINEEND, ENDMARKER


@pytest.mark.parametrize(
    "string,expected",
    (
        (
            "# comment",
            [
                COMMENT(1, 1, "# comment"),
                IMPLICITLINEEND(1, 10),
                ENDMARKER(2),
            ],
        ),
        (
            "### comment",
            [
                COMMENT(1, 1, "### comment"),
                IMPLICITLINEEND(1, 12),
                ENDMARKER(2),
            ],
        ),
        (
            "### comment ###",
            [
                COMMENT(1, 1, "### comment ###"),
                IMPLICITLINEEND(1, 16),
                ENDMARKER(2),
            ],
        ),
    ),
)
def test_comment(tokenizer, string, expected):
    assert list(tokenizer.tokenize(string)) == expected



"""Loging utilities."""

import logging
from contextlib import contextmanager
from fnmatch import fnmatch


@contextmanager
def logs(logger, level=None, handler=None, close=True):
    """Emit logs at or above `level` in the encapsulated context, optionally using the
    specified `handler`.

    See `the Python logging cookbook
    <https://docs.python.org/3/howto/logging-cookbook.html#using-a-context-manager-for-selective-logging>`__
    for more information.

    Parameters
    ----------
    logger : :class:`logging.Logger`
        The logger to use for the encapsulated context.

    level : str or int, optional
        The minimum log levels to emit. The standard log levels "debug", "info",
        "warning", "error" and "critical" are supported, as are their corresponding
        level numbers (see :mod:`logging`).

    handler : :class:`logging.Handler`, optional
        The handler to add to `logger` for the encapsulated context.

    close : bool, optional
        Close `handler` once finished. Defaults to `True`.

    Examples
    --------
    Print debug logs during parsing, excluding compilation messages.

    >>> import logging
    >>> from finesse import Model
    >>> from finesse.utilities import logs
    >>> from finesse.utilities.logging import FinesseStreamHandler
    >>> handler = FinesseStreamHandler()
    >>> handler.exclude("finesse.script.compiler")
    >>> model = Model()
    >>> with logs(logging.getLogger(), level="debug", handler=handler):
    >>>     model.parse("laser l1 P=1")
    """
    old_level = None

    if level is not None:
        try:
            # Convert level name to uppercase.
            level = level.upper()
        except AttributeError:
            # Probably a number.
            pass

        old_level = logger.level
        logger.setLevel(level)

    if handler:
        logger.addHandler(handler)

    yield

    if level is not None:
        logger.setLevel(old_level)

    if handler:
        logger.removeHandler(handler)

        if close:
            handler.close()


@contextmanager
def tracebacks(tracebacks=True):
    """Show or hide tracebacks in the encapsulated context.

    Some environments, such as Jupyterlab, hide tracebacks by default. This context
    allows tracebacks to be forcefully shown or hidden temporarily.

    Parameters
    ----------
    tracebacks : bool, optional
        Show tracebacks. Defaults to True.

    Examples
    --------
    Print tracebacks during a run.

    >>> from finesse import Model
    >>> from finesse.utilities import tracebacks
    >>> model = Model()
    >>> with tracebacks():
    >>>     model.parse("laser l1 L=1")
    """
    from .. import show_tracebacks

    old_tracebacks = show_tracebacks(tracebacks)
    yield
    show_tracebacks(old_tracebacks)


class FinesseStreamHandler(logging.StreamHandler):
    """Finesse stream handler.

    This class provides a mechanism to exclude displayed log channels by wildcard. It is
    otherwise identical to :class:`logging.StreamHandler`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__excluded_channels = None
        self.reset_exclude_patterns()

    def exclude(self, pattern):
        self.__excluded_channels.add(pattern)

    def reset_exclude_patterns(self):
        """Empty the configured log channel exclude patterns, and return what was
        there."""
        old_excludes = set(self.__excluded_channels or [])
        self.__excluded_channels = set()
        return old_excludes

    def filter(self, record):
        for pattern in self.__excluded_channels:
            if fnmatch(record.name, pattern):
                # Skip the record.
                return
        return record



"""Controlling an interferometer via error signals."""

import logging

from finesse.parameter import ParameterRef, bool_parameter, float_parameter
from finesse.element import ModelElement

LOGGER = logging.getLogger(__name__)


@float_parameter("gain", "Gain of the lock")
@float_parameter("accuracy", "Accuracy to reach for the lock")
@float_parameter("offset", "Offset applied to the input of the error signal")
@bool_parameter("enabled", "Whether the lock is enabled or not")
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Lock(ModelElement):
    """A simple lock which computes and applies the feedback to a given parameter using
    an error signal.

    Parameters
    ----------
    name : str
        Name of newly created lock.

    error_signal : Any
        An error signal parameter or an object capable of producing a real-type error signal. This
        is typically a demodulated :class:`.PowerDetector` instance (or the name of the instance).

    feedback : :class:`.Parameter`
        A parameter of the model to apply the locks' feedback signal to.

    gain : float
        Control loop gain.

    accuracy : float
        Threshold to decide whether the loop is locked.

    enabled : boolean
        If true this lock will run when the `RunLocks()` action is used. Explicitly specifying
        the name of the lock will override this setting, e.g. `RunLocks(name)`.

    offset : float
        An offset that is applied to the error signal before it is used.
    """

    def __init__(
        self, name, error_signal, feedback, gain, accuracy, *, enabled=True, offset=0
    ):
        super().__init__(name)

        self.enabled = enabled
        self.__errsig = error_signal
        self.__feedback = feedback
        self.gain = gain
        self.accuracy = accuracy
        self.offset = offset

    def _on_add(self, model):
        if isinstance(self.__errsig, str):
            self.__errsig = model.elements[self.__errsig]

    @property
    def disabled(self):
        return not self.enabled.value

    @property
    def error_signal(self):
        """The error signal of the lock."""
        return self.__errsig

    @error_signal.setter
    def error_signal(self, value):
        self.__errsig = self._model.get_element(value)

    @property
    def feedback(self):
        """A handle to the parameter which the feedback signal is applied to."""
        if isinstance(self.__feedback, ParameterRef):
            return self.__feedback.parameter
        else:
            return self.__feedback

    @feedback.setter
    def feedback(self, value):
        self.__feedback = value

    def __repr__(self):
        return f"<'{self.name}' @ {hex(id(self))} ({self.__class__.__name__}) enabled={self.enabled.value}>"



import pytest
from testutils.text import dedent_multiline


# Override default element to take four arguments.
@pytest.fixture
def fake_element_cls(fake_element_cls):
    class FakeElement(fake_element_cls):
        def __init__(self, name, a, b, c, d):
            super().__init__(name)
            self.a = a
            self.b = b
            self.c = c
            self.d = d

    return FakeElement


@pytest.fixture
def spec(spec, set_spec_constructs, fake_element_adapter_factory, fake_element_cls):
    spec.register_element(fake_element_adapter_factory(fake_element_cls))
    set_spec_constructs("keywords", {"E", "F", "G", "H"})

    return spec


@pytest.mark.parametrize(
    "script,graph_definition",
    (
        pytest.param(
            "fake_element myelement E F G H",
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label "kat"
                        type "ROOT"
                        extra_tokens "[]"
                    ]
                    node [
                        id 1
                        label "kat.0"
                        type "ELEMENT"
                        token "'fake_element'"
                        name_token "'myelement'"
                        extra_tokens "[' ', ' ', ' ', ' ', ' ']"
                    ]
                    node [
                        id 2
                        label "kat.0.0"
                        token "'E'"
                        type "KEYWORD"
                    ]
                    node [
                        id 3
                        label "kat.0.1"
                        token "'F'"
                        type "KEYWORD"
                    ]
                    node [
                        id 4
                        label "kat.0.2"
                        token "'G'"
                        type "KEYWORD"
                    ]
                    node [
                        id 5
                        label "kat.0.3"
                        token "'H'"
                        type "KEYWORD"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 1
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 3
                        target 1
                        type "ARGUMENT"
                        order "1"
                    ]
                    edge [
                        source 4
                        target 1
                        type "ARGUMENT"
                        order "2"
                    ]
                    edge [
                        source 5
                        target 1
                        type "ARGUMENT"
                        order "3"
                    ]
                ]
                """
            ),
            id="element with 4 keywords",
        ),
    ),
)
def test_keywords(assert_graphs_match, script, graph_definition):
    assert_graphs_match(script, graph_definition)



"""Parameter unit tests."""

from finesse.symbols import Constant
from finesse.parameter import int_parameter, bool_parameter
from finesse.element import ModelElement


def test_parameter_equality_checks(model):
    """Test that parameters can be directly compared to numbers and other symbols."""
    model.parse("laser l1 P=3.14")
    assert model.l1.P == 3.14
    assert model.l1.P == Constant(3.14)
    assert model.l1.P == (Constant(3.14) + Constant(3.14)) / 2
    assert model.l1.P.ref == model.l1.P


def test_parameter_ref_substitution(model):
    """Test that parameter refs can be subsituted for in expressions."""
    model.parse("laser l1 P=3.14")
    assert model.l1.P.eval() == 3.14
    assert (model.l1.P.ref + 1).eval(subs={model.l1.P: 4}) == 5


def test_bool_parameter_checks(model):
    """Test bool parameters get boolean checked properly."""

    @bool_parameter("bool", "")
    class TEST(ModelElement):
        pass

    model.add(TEST("test"))
    model.test.bool = True
    assert model.test.bool
    assert model.test.bool.ref
    model.test.bool = False
    assert not model.test.bool
    assert not model.test.bool.ref


def test_parameter_table(model):
    @int_parameter("test_param", "number", "Hz")
    class TEST(ModelElement):
        pass

    model.add(TEST("test1"))
    model.add(TEST("test2"))
    model.test1.test_param = 10
    model.test2.test_param = model.test1.test_param.ref
    assert model.test1.info() == (
        "TEST test1\n\n"
        "Parameters:\n"
        "┌─────────────┬───────┐\n"
        "│ Description │ Value │\n"
        "╞═════════════╪═══════╡\n"
        "│ number      │ 10 Hz │\n"
        "└─────────────┴───────┘\n"
    )
    assert model.test2.info() == (
        "TEST test2\n\n"
        "Parameters:\n"
        "┌─────────────┬─────────────────────┐\n"
        "│ Description │ Value               │\n"
        "╞═════════════╪═════════════════════╡\n"
        "│ number      │ test1.test_param Hz │\n"
        "└─────────────┴─────────────────────┘\n"
    )
    assert model.test2.info(eval_refs=True) == (
        "TEST test2\n\n"
        "Parameters:\n"
        "┌─────────────┬───────┐\n"
        "│ Description │ Value │\n"
        "╞═════════════╪═══════╡\n"
        "│ number      │ 10 Hz │\n"
        "└─────────────┴───────┘\n"
    )


def test_parameter_ref_string_repr(model):
    "Test functionality for resolving refs when printing parameter values"
    model.parse("laser l1 P=3.14")
    model.parse("laser l2 l1.P")
    assert str(model.l1.P) == "3.14 W"
    assert str(model.l2.P) == "l1.P W"
    model.l2.P.eval_string = True
    assert str(model.l2.P) == "3.14 W"



"""Tests for the phase of the light field as propagated over a space."""

import pytest
import numpy as np

from finesse.analysis.actions import Noxaxis, Xaxis
import finesse.constants as constants
from finesse.components import Laser, Mirror
from finesse.detectors import AmplitudeDetector


@pytest.fixture
def laser_and_mirror_with_ad(model):
    """Model with a laser and mirror separated by a space, with amplitude detector."""
    model.chain(Laser("L0"), {"name": "s0"}, Mirror("M", R=1, T=0))
    model.add(AmplitudeDetector("ad_lsr", model.L0.p1.o, model.L0.f.ref))
    model.add(AmplitudeDetector("ad_in", model.M.p1.i, model.L0.f.ref))
    model.add(AmplitudeDetector("ad_refl", model.M.p1.o, model.L0.f.ref))
    model.add(AmplitudeDetector("ad_back", model.L0.p1.i, model.L0.f.ref))
    return model


@pytest.mark.parametrize("L", (0, 1, 3.141, 1000, 3141))
def test_propagation__carrier(laser_and_mirror_with_ad, L):
    """Test carrier amplitude doesn't change with 0-length propagation."""
    laser_and_mirror_with_ad.spaces.s0.L = L
    out = laser_and_mirror_with_ad.run(Noxaxis())
    assert out["ad_lsr"] == out["ad_in"]
    assert out["ad_lsr"] == out["ad_refl"]
    assert out["ad_lsr"] == out["ad_back"]


@pytest.mark.parametrize("L", (0, 1, 3.141, 1000, 3141))
def test_propagation__frequency_offset(laser_and_mirror_with_ad, L):
    """Test frequency offset amplitudes don't change with 0-length propagation."""
    laser_and_mirror_with_ad.spaces.s0.L = L

    freqs = np.linspace(-1e3, 1e3, 100)

    out = laser_and_mirror_with_ad.run(
        Xaxis("L0.f", "lin", freqs[0], freqs[-1], freqs.size - 1)
    )

    # Assert that field at L0.p1.o doesn't change.
    assert np.allclose(
        np.diff(out["ad_lsr"]), np.zeros(freqs.size - 1, dtype=np.complex128)
    )

    s0_L = laser_and_mirror_with_ad.elements["s0"].L.value
    s0_n = laser_and_mirror_with_ad.elements["s0"].nr.value

    # Test that the field propagation across the space (towards mirror M) is equal to expected
    # analytic result.
    prop = np.exp(-1j * 2 * np.pi * freqs * s0_L * s0_n / constants.C_LIGHT)
    assert np.allclose(out["ad_in"], prop)

    # And test that the field propagation across the space in the other direction (back towards
    # laser L0) is equal to just another factor of the propagation term from above.
    assert np.allclose(out["ad_back"], prop**2)



"""
Physical constants module exposing a single named-tuple
which contains each of the constants values.

This tuple is imported in the top-level Finesse module - allowing
access to this structure with::

    from finesse import constants

    # get the speed of light for example
    c = constants.C_LIGHT

The table below lists all of the constants exposed in this module.

+-----------+------------------------------------+------------------------+
| Name      | Description                        | Exact value used       |
+===========+====================================+========================+
| `C_LIGHT` | The speed of light in m/s.         | 299792458.0            |
+-----------+------------------------------------+------------------------+
| `PI`      | Mathematical constant :math:`\pi`. | 3.14159265358979323846 |
+-----------+------------------------------------+------------------------+

"""
from collections import namedtuple
values = namedtuple("Constants", "PI C_LIGHT RAD2DEG DEG2RAD H_PLANCK")(_PI, _C_LIGHT, _RAD2DEG, _DEG2RAD, _H_PLANCK)



"""Shared utilities for Finesse tests."""

from finesse.script.spec import KATSPEC


# Supported KatScript operators.
BINARY_OPERATORS = KATSPEC.binary_operators
UNARY_OPERATORS = KATSPEC.unary_operators
CONSTANTS = KATSPEC.constants
EXPRESSION_FUNCTIONS = KATSPEC.expression_functions
